"""
This module provides functions for correcting the a/e energy dependence, determining the cut level and calculating survival fractions.
"""

from __future__ import annotations

import json
import logging
import os
import pathlib

import matplotlib as mpl

mpl.use("agg")
import matplotlib.cm as cmx
import matplotlib.colors as mcolors
import matplotlib.pyplot as plt
import numpy as np
from iminuit import Minuit, cost, util
from matplotlib.backends.backend_pdf import PdfPages
from matplotlib.colors import LogNorm
from scipy.stats import chi2

import pygama.lgdo.lh5_store as lh5
import pygama.math.histogram as pgh
import pygama.math.peak_fitting as pgf
import pygama.pargen.ecal_th as thc
import pygama.pargen.energy_cal as pgc

log = logging.getLogger(__name__)


def load_aoe(
    files: list,
    lh5_path: str,
    cal_dict: dict,
    energy_param: str,
    cal_energy_param: str,
    cut_field: str = "Cal_cuts",
) -> tuple(np.array, np.array, np.array, np.array):
    """
    Loads in the A/E parameters needed and applies calibration constants to energy
    """

    sto = lh5.LH5Store()

    params = ["A_max", "tp_0_est", "tp_99", "dt_eff", energy_param, cal_energy_param]

    table = sto.read_object(lh5_path, files)[0]
    df = table.eval(cal_dict).get_dataframe()

    param_dict = {}
    for param in params:
        # add cuts in here
        if param in df:
            param_dict[param] = df[param].to_numpy()
        else:
            param_dict.update(lh5.load_nda(files, [param], lh5_path))
    if cut_field in df.keys():
        for entry in param_dict:
            param_dict[entry] = param_dict[entry][df[cut_field].to_numpy()]

    aoe = np.divide(param_dict["A_max"], param_dict[energy_param])
    return aoe, param_dict[cal_energy_param], param_dict["dt_eff"]


def PDF_AoE(
    x: np.array,
    lambda_s: float,
    lambda_b: float,
    mu: float,
    sigma: float,
    tau: float,
    lower_range: float = np.inf,
    upper_range: float = np.inf,
    components: bool = False,
) -> tuple(float, np.array):
    """
    PDF for A/E consists of a gaussian signal with gaussian tail background
    """
    try:
        sig = lambda_s * pgf.gauss_norm(x, mu, sigma)
        bkg = lambda_b * pgf.gauss_tail_norm(
            x, mu, sigma, tau, lower_range, upper_range
        )
    except:
        sig = np.full_like(x, np.nan)
        bkg = np.full_like(x, np.nan)

    if components == False:
        pdf = sig + bkg
        return lambda_s + lambda_b, pdf
    else:
        return lambda_s + lambda_b, sig, bkg


def unbinned_aoe_fit(
    aoe: np.array, display: int = 0, verbose: bool = False
) -> tuple(np.array, np.array):
    """
    Fitting function for A/E, first fits just a gaussian before using the full pdf to fit
    if fails will return NaN values
    """

    hist, bins, var = pgh.get_hist(aoe, bins=500)
    bin_centers = (bins[:-1] + bins[1:]) / 2

    pars, cov = pgf.gauss_mode_max(hist, bins)
    mu = bin_centers[np.argmax(hist)]
    _, sigma, _ = pgh.get_gaussian_guess(hist, bins)
    ls_guess = 2 * np.sum(hist[(bin_centers > mu) & (bin_centers < (mu + 2.5 * sigma))])
    c1_min = mu - 2 * sigma
    c1_max = mu + 5 * sigma

    # Initial fit just using Gaussian
    c1 = cost.UnbinnedNLL(aoe[(aoe < c1_max) & (aoe > c1_min)], pgf.gauss_pdf)
    m1 = Minuit(c1, mu, sigma, ls_guess)
    m1.limits = [
        (mu * 0.8, mu * 1.2),
        (0.8 * sigma, sigma * 1.2),
        (0, len(aoe[(aoe < c1_max) & (aoe > c1_min)])),
    ]
    m1.migrad()
    mu, sigma, ls_guess = m1.values
    if verbose:
        print(m1)

    # Range to fit over, below this tail behaviour more exponential, few events above
    fmin = mu - 15 * sigma
    fmax = mu + 5 * sigma

    bg_guess = len(aoe[(aoe < fmax) & (aoe > fmin)]) - ls_guess
    x0 = [ls_guess, bg_guess, mu, sigma, 0.01, fmin, fmax, 0]
    if verbose:
        print(x0)

    # Full fit using gaussian signal with gaussian tail background
    c = cost.ExtendedUnbinnedNLL(aoe[(aoe < fmax) & (aoe > fmin)], PDF_AoE)
    m = Minuit(c, *x0)
    m.fixed[5:] = True
    m.simplex().migrad()
    m.hesse()
    if verbose:
        print(m)

    valid = m.valid & m.accurate
    if valid == False:
        return np.full_like(x0, np.nan), np.full_like(x0, np.nan)

    if display > 1:
        plt.figure()
        xs = np.linspace(fmin, fmax, 1000)
        counts, bins, bars = plt.hist(
            aoe[(aoe < fmax) & (aoe > fmin)], bins=400, histtype="step", label="Data"
        )
        dx = np.diff(bins)
        plt.plot(xs, PDF_AoE(xs, *m.values)[1] * dx[0], label="Full fit")
        n_events, sig, bkg = PDF_AoE(xs, *m.values[:-1], True)
        plt.plot(xs, sig * dx[0], label="Signal")
        plt.plot(xs, bkg * dx[0], label="Background")
        plt.plot(xs, pgf.gauss_pdf(xs, *m1.values) * dx[0], label="Initial Gaussian")
        plt.legend(loc="upper left")
        plt.show()

        plt.figure()
        bin_centers = (bins[1:] + bins[:-1]) / 2
        res = (PDF_AoE(bin_centers, *m.values)[1] * dx[0]) - counts
        plt.plot(
            bin_centers,
            [re / count if count != 0 else re for re, count in zip(res, counts)],
            label="Normalised Residuals",
        )
        plt.legend(loc="upper left")
        plt.show()
        return m.values, m.errors

    else:
        return m.values, m.errors


def pol1(x: np.array, a: float, b: float) -> np.array:
    """Basic Polynomial for fitting A/E centroid against energy"""
    return a * x + b


def sigma_fit(x: np.array, a: float, b: float) -> np.array:
    """Function definition for fitting A/E sigma against energy"""
    return np.sqrt(a + (b / x) ** 2)


def AoEcorrection(
    energy: np.array, aoe: np.array, eres: list, plot_dict: dict = {}, display: int = 0
) -> tuple(np.array, np.array):
    """
    Calculates the corrections needed for the energy dependence of the A/E.
    Does this by fitting the compton continuum in slices and then applies fits to the centroid and variance.
    """

    comptBands_width = 20
    comptBands = np.array(
        [
            940,
            960,
            980,
            1000,
            1020,
            1040,
            1130,
            1150,
            1170,
            1190,
            1210,
            1250,
            1270,
            1290,
            1310,
            1330,
            1370,
            1390,
            1420,
            1520,
            1540,
            1650,
            1700,
            1780,
            1810,
            1850,
            1870,
            1890,
            1910,
            1930,
            1950,
            1970,
            1990,
            2010,
            2030,
            2050,
            2150,
            2170,
            2190,
            2210,
            2230,
            2250,
            2270,
            2290,
            2310,
            2330,
            2350,
        ]
    )
    results_dict = {}
    comptBands = comptBands[::-1]  # Flip so color gets darker when plotting
    # peaks = np.array([1080,1094,1459,1512, 1552, 1592,1620, 1650, 1670,1830,2105])
    compt_aoe = np.zeros(len(comptBands))
    aoe_sigmas = np.zeros(len(comptBands))
    compt_aoe_err = np.zeros(len(comptBands))
    aoe_sigmas_err = np.zeros(len(comptBands))
    ratio = np.zeros(len(comptBands))
    ratio_err = np.zeros(len(comptBands))

    copper = cm = plt.get_cmap("copper")
    cNorm = mcolors.Normalize(vmin=0, vmax=len(comptBands))
    scalarMap = cmx.ScalarMappable(norm=cNorm, cmap=copper)

    if display > 0:
        fits_fig = plt.figure()

    # Fit each compton band
    for i, band in enumerate(comptBands):
        aoe_tmp = aoe[
            (energy > band) & (energy < band + comptBands_width) & (aoe > 0)
        ]  # [:20000]
        try:
            pars, errs = unbinned_aoe_fit(aoe_tmp, display=display)
            compt_aoe[i] = pars[2]
            aoe_sigmas[i] = pars[3]
            compt_aoe_err[i] = errs[2]
            aoe_sigmas_err[i] = errs[3]
            ratio[i] = pars[0] / pars[1]
            ratio_err[i] = ratio[i] * np.sqrt(
                (errs[0] / pars[0]) ** 2 + (errs[1] / pars[1]) ** 2
            )
        except:
            compt_aoe[i] = np.nan
            aoe_sigmas[i] = np.nan
            compt_aoe_err[i] = np.nan
            aoe_sigmas_err[i] = np.nan
            ratio[i] = np.nan
            ratio_err[i] = np.nan

        if display > 0:
            if np.isnan(errs[2]) | np.isnan(errs[3]) | (errs[2] == 0) | (errs[3] == 0):
                pass
            else:
                xs = np.arange(
                    pars[2] - 4 * pars[3], pars[2] + 3 * pars[3], pars[3] / 10
                )
                colorVal = scalarMap.to_rgba(i)
                plt.plot(xs, PDF_AoE(xs, *pars)[1], color=colorVal)

    if display > 0:
        plt.xlabel("A/E")
        plt.ylabel("Expected Counts")
        plt.title("Compton Band Fits")
        cbar = plt.colorbar(
            cmx.ScalarMappable(norm=cNorm, cmap=plt.get_cmap("copper_r")),
            orientation="horizontal",
            label="Compton Band Energy",
            ticks=[0, 16, 32, len(comptBands)],
        )  # cax=ax,
        cbar.ax.set_xticklabels(
            [
                comptBands[::-1][0],
                comptBands[::-1][16],
                comptBands[::-1][32],
                comptBands[::-1][-1],
            ]
        )
        plot_dict["band_fits"] = fits_fig
        if display > 1:
            plt.show()
        else:
            plt.close()

    ids = (
        np.isnan(compt_aoe_err)
        | np.isnan(aoe_sigmas_err)
        | (aoe_sigmas_err == 0)
        | (compt_aoe_err == 0)
    )
    results_dict["n_of_valid_fits"] = len(np.where(~ids)[0])
    # Fit mus against energy
    p0_mu = [-1e-06, 5e-01]
    c_mu = cost.LeastSquares(
        comptBands[~ids], compt_aoe[~ids], compt_aoe_err[~ids], pol1
    )
    c_mu.loss = "soft_l1"
    m_mu = Minuit(c_mu, *p0_mu)
    m_mu.simplex()
    m_mu.migrad()
    m_mu.hesse()

    pars = m_mu.values
    errs = m_mu.errors

    csqr_mu = np.sum(
        ((compt_aoe[~ids] - pol1(comptBands[~ids], *pars)) ** 2) / compt_aoe_err[~ids]
    )
    dof_mu = len(compt_aoe[~ids]) - len(pars)
    results_dict["p_val_mu"] = chi2.sf(csqr_mu, dof_mu)
    results_dict["csqr_mu"] = (csqr_mu, dof_mu)

    # Fit sigma against energy
    p0_sig = [np.nanpercentile(aoe_sigmas[~ids], 50) ** 2, 2]
    c_sig = cost.LeastSquares(
        comptBands[~ids], aoe_sigmas[~ids], aoe_sigmas_err[~ids], sigma_fit
    )
    c_sig.loss = "soft_l1"
    m_sig = Minuit(c_sig, *p0_sig)
    m_sig.simplex()
    m_sig.migrad()
    m_sig.hesse()

    sig_pars = m_sig.values
    sig_errs = m_sig.errors

    csqr_sig = np.sum(
        ((aoe_sigmas[~ids] - sigma_fit(comptBands[~ids], *sig_pars)) ** 2)
        / aoe_sigmas_err[~ids]
    )
    dof_sig = len(aoe_sigmas[~ids]) - len(sig_pars)
    results_dict["p_val_sig"] = chi2.sf(csqr_sig, dof_sig)
    results_dict["csqr_sig"] = (csqr_sig, dof_sig)

    model = pol1(comptBands, *pars)
    sig_model = sigma_fit(comptBands, *sig_pars)

    # Get DEP fit
    sigma = np.sqrt(eres[0] + 1592 * eres[1]) / 2.355
    n_sigma = 4
    peak = 1592
    emin = peak - n_sigma * sigma
    emax = peak + n_sigma * sigma
    dep_pars, dep_err = unbinned_aoe_fit(
        aoe[(energy > emin) & (energy < emax) & (aoe > 0)][:10000]
    )

    if display > 0:
        mean_fig, (ax1, ax2) = plt.subplots(2, 1, sharex=True)
        ax1.errorbar(
            comptBands[~ids] + 10,
            compt_aoe[~ids],
            yerr=compt_aoe_err[~ids],
            xerr=10,
            label="data",
            linestyle=" ",
        )
        ax1.plot(comptBands[~ids] + 10, model[~ids], label="linear model")
        ax1.errorbar(
            1592,
            dep_pars[2],
            xerr=n_sigma * sigma,
            yerr=dep_err[2],
            label="DEP",
            color="green",
            linestyle=" ",
        )

        ax1.legend(title="A/E mu energy dependence", frameon=False)

        ax1.set_ylabel("raw A/E (a.u.)", ha="right", y=1)
        ax2.scatter(
            comptBands[~ids] + 10,
            100 * (compt_aoe[~ids] - model[~ids]) / compt_aoe_err[~ids],
            lw=1,
            c="b",
        )
        ax2.scatter(
            1592, 100 * (dep_pars[2] - pol1(1592, *pars)) / dep_err[2], lw=1, c="g"
        )
        ax2.set_ylabel("Residuals %", ha="right", y=1)
        ax2.set_xlabel("Energy (keV)", ha="right", x=1)
        plt.tight_layout()
        plot_dict["mean_fit"] = mean_fig
        if display > 1:
            plt.show()
        else:
            plt.close()

    if display > 0:
        sig_fig, (ax1, ax2) = plt.subplots(2, 1, sharex=True)
        ax1.errorbar(
            comptBands[~ids] + 10,
            aoe_sigmas[~ids],
            yerr=aoe_sigmas_err[~ids],
            xerr=10,
            label="data",
            linestyle=" ",
        )
        ax1.plot(
            comptBands[~ids],
            sig_model[~ids],
            label=f"sqrt model: sqrt({sig_pars[0]:1.4f}+({sig_pars[1]:1.1f}/E)^2)",
        )  # {sig_pars[2]:1.1f}
        ax1.errorbar(
            1592,
            dep_pars[3],
            xerr=n_sigma * sigma,
            yerr=dep_err[3],
            label="DEP",
            color="green",
        )
        ax1.set_ylabel("A/E stdev (a.u.)", ha="right", y=1)
        ax1.legend(title="A/E stdev energy dependence", frameon=False)
        ax2.scatter(
            comptBands[~ids] + 10,
            100 * (aoe_sigmas[~ids] - sig_model[~ids]) / aoe_sigmas_err[~ids],
            lw=1,
            c="b",
        )
        ax2.scatter(
            1592,
            100 * (dep_pars[3] - sigma_fit(1592, *sig_pars)) / dep_err[3],
            lw=1,
            c="g",
        )
        ax2.set_ylabel("Residuals", ha="right", y=1)
        ax2.set_xlabel("Energy (keV)", ha="right", x=1)
        plt.tight_layout()
        plot_dict["sigma_fit"] = sig_fig
        if display > 1:
            plt.show()
        else:
            plt.close()
        return pars, sig_pars, results_dict, plot_dict
    else:
        return pars, sig_pars, results_dict


def plot_compt_bands_overlayed(
    aoe: np.array, energy: np.array, eranges: list[tuple], aoe_range: list[float] = None
) -> None:
    """
    Function to plot various compton bands to check energy dependence and corrections
    """

    for erange in eranges:
        hist, bins, var = pgh.get_hist(
            aoe[(energy > erange - 10) & (energy < erange + 10) & (~np.isnan(aoe))],
            bins=500,
        )
        bin_cs = (bins[1:] + bins[:-1]) / 2
        mu = bin_cs[np.argmax(hist)]
        if aoe_range is None:
            aoe_range = [mu * 0.97, mu * 1.02]
        idxs = (
            (energy > erange - 10)
            & (energy < erange + 10)
            & (aoe > aoe_range[0])
            & (aoe < aoe_range[1])
        )
        plt.hist(aoe[idxs], bins=50, histtype="step", label=f"{erange-10}-{erange+10}")


def plot_dt_dep(
    aoe: np.array, energy: np.array, dt: np.array, erange: list[tuple], title: str
) -> None:
    """
    Function to produce 2d histograms of A/E against drift time to check dependencies
    """

    hist, bins, var = pgh.get_hist(
        aoe[(energy > erange[0]) & (energy < erange[1]) & (~np.isnan(aoe))], bins=500
    )
    bin_cs = (bins[1:] + bins[:-1]) / 2
    mu = bin_cs[np.argmax(hist)]
    aoe_range = [mu * 0.9, mu * 1.1]

    idxs = (
        (energy > erange[0])
        & (energy < erange[1])
        & (aoe > aoe_range[0])
        & (aoe < aoe_range[1])
        & (dt < 2000)
    )

    plt.hist2d(aoe[idxs], dt[idxs], bins=[200, 100], norm=LogNorm())
    plt.ylabel("Drift Time (ns)")
    plt.xlabel("A/E")
    plt.title(title)


def energy_guess(hist, bins, var, func_i, peak, eres_pars, fit_range):
    """
    Simple guess for peak fitting
    """
    if func_i == pgf.extended_radford_pdf:
        bin_cs = (bins[1:] + bins[:-1]) / 2
        sigma = thc.fwhm_slope(peak, *eres_pars) / 2.355
        i_0 = np.nanargmax(hist)
        mu = peak
        height = hist[i_0]
        bg0 = np.mean(hist[-10:])
        step = np.mean(hist[:10]) - bg0
        htail = 1.0 / 5
        tau = 0.5 * sigma

        hstep = step / (bg0 + np.mean(hist[:10]))
        dx = np.diff(bins)[0]
        n_bins_range = int((3 * sigma) // dx)
        nsig_guess = np.sum(hist[i_0 - n_bins_range : i_0 + n_bins_range])
        nbkg_guess = np.sum(hist) - nsig_guess
        if nbkg_guess < 0:
            nbkg_guess = 0
        if nsig_guess < 0:
            nsig_guess = 0
        parguess = [
            nsig_guess,
            mu,
            sigma,
            htail,
            tau,
            nbkg_guess,
            hstep,
            fit_range[0],
            fit_range[1],
            0,
        ]  #
        return parguess

    elif func_i == pgf.extended_gauss_step_pdf:
        mu = peak
        sigma = thc.fwhm_slope(peak, *eres_pars) / 2.355
        i_0 = np.argmax(hist)
        bg = np.mean(hist[-10:])
        step = bg - np.mean(hist[:10])
        hstep = step / (bg + np.mean(hist[:10]))
        dx = np.diff(bins)[0]
        n_bins_range = int((3 * sigma) // dx)
        nsig_guess = np.sum(hist[i_0 - n_bins_range : i_0 + n_bins_range])
        nbkg_guess = np.sum(hist) - nsig_guess
        if nbkg_guess < 0:
            nbkg_guess = 0
        if nsig_guess < 0:
            nsig_guess = 0
        return [nsig_guess, mu, sigma, nbkg_guess, hstep, fit_range[0], fit_range[1], 0]


def unbinned_energy_fit(
    energy: np.array,
    peak: float,
    eres_pars: list = None,
    simplex=False,
    guess=None,
    verbose: bool = False,
) -> tuple(np.array, np.array):
    """
    Fitting function for energy peaks used to calculate survival fractions
    """
    hist, bins, var = pgh.get_hist(
        energy, dx=0.5, range=(np.nanmin(energy), np.nanmax(energy))
    )
    if guess is None:
        x0 = energy_guess(
            hist,
            bins,
            var,
            pgf.extended_gauss_step_pdf,
            peak,
            eres_pars,
            (np.nanmin(energy), np.nanmax(energy)),
        )
        c = cost.ExtendedUnbinnedNLL(energy, pgf.extended_gauss_step_pdf)
        m = Minuit(c, *x0)
        m.fixed[-3:] = True
        m.simplex().migrad()
        m.hesse()
        x0 = m.values[:3]
        x0 += [1 / 5, 0.2 * m.values[2]]
        x0 += m.values[3:]
        if verbose:
            print(m)
    else:
        x0 = guess
    if len(x0) == 0:
        return [np.nan], [np.nan]

    fixed, mask = pgc.get_hpge_E_fixed(pgf.extended_radford_pdf)
    if verbose:
        print(x0)
    c = cost.ExtendedUnbinnedNLL(energy, pgf.extended_radford_pdf)
    m = Minuit(c, *x0)
    if x0[0] <= 0:
        sig_lim = 0.2 * len(energy)
    else:
        sig_lim = 1.2 * x0[0]
    m.limits = [
        (0, sig_lim),
        (None, None),
        (None, None),
        (0, 1),
        (0, None),
        (0, len(energy) * 1.1),
        (None, None),
        (None, None),
        (None, None),
        (None, None),
    ]
    for fix in fixed:
        m.fixed[fix] = True
    if simplex == True:
        m.simplex().migrad()
    else:
        m.migrad()

    m.hesse()
    if verbose:
        print(m)

    valid = m.valid

    if valid == True and not np.isnan(m.errors[:-3]).all():
        return m.values, m.errors
    else:
        x0 = energy_guess(
            hist,
            bins,
            var,
            pgf.extended_radford_pdf,
            peak,
            eres_pars,
            (np.nanmin(energy), np.nanmax(energy)),
        )
        c = cost.ExtendedUnbinnedNLL(energy, pgf.extended_radford_pdf)
        m = Minuit(c, *x0)
        if x0[0] <= 0:
            sig_lim = 0.2 * len(energy)
        else:
            sig_lim = 1.2 * x0[0]
        m.limits = [
            (0, sig_lim),
            (None, None),
            (None, None),
            (0, 1),
            (0, None),
            (0, len(energy) * 1.1),
            (None, None),
            (None, None),
            (None, None),
            (None, None),
        ]
        for fix in fixed:
            m.fixed[fix] = True
        m.simplex().migrad()
        m.hesse()
        valid = m.valid
        if verbose:
            print(m)
        if valid == True:
            return m.values, m.errors
        else:
            return np.full_like(x0, np.nan), np.full_like(x0, np.nan)


def get_peak_label(peak: float) -> str:
    if peak == 2039:
        return "CC @"
    elif peak == 1592.5:
        return "Tl DEP @"
    elif peak == 1620.5:
        return "Bi FEP @"
    elif peak == 2103.53:
        return "Tl SEP @"
    elif peak == 2614.5:
        return "Tl FEP @"


def get_survival_fraction(
    energy,
    aoe,
    cut_val,
    peak,
    eres_pars,
    high_cut=None,
    guess_pars_cut=None,
    guess_pars_surv=None,
):
    if high_cut is not None:
        idxs = (aoe > cut_val) & (aoe < high_cut)
    else:
        idxs = aoe > cut_val
    cut_pars, ct_errs = unbinned_energy_fit(
        energy[~idxs], peak, eres_pars, guess=guess_pars_cut, simplex=True
    )
    surv_pars, surv_errs = unbinned_energy_fit(
        energy[idxs], peak, eres_pars, guess=guess_pars_surv, simplex=True
    )

    ct_n = cut_pars[0]
    ct_err = ct_errs[0]
    surv_n = surv_pars[0]
    surv_err = surv_errs[0]

    pc_n = ct_n + surv_n
    pc_err = np.sqrt(surv_err**2 + ct_err**2)

    sf = (surv_n / pc_n) * 100
    err = sf * np.sqrt((pc_err / pc_n) ** 2 + (surv_err / surv_n) ** 2)
    return sf, err, cut_pars, surv_pars


def get_aoe_cut_fit(
    energy: np.array,
    aoe: np.array,
    peak: float,
    ranges: tuple(int, int),
    dep_acc: float,
    eres_pars: list,
    display: int = 1,
) -> float:
    """
    Determines A/E cut by sweeping through values and for each one fitting the DEP to determine how many events survive.
    Then interpolates to get cut value at desired DEP survival fraction (typically 90%)
    """

    min_range, max_range = ranges

    peak_energy = energy[(energy > peak - min_range) & (energy < peak + max_range)][
        :20000
    ]
    peak_aoe = aoe[(energy > peak - min_range) & (energy < peak + max_range)][:20000]

    cut_vals = np.arange(-8, 0, 0.2)
    pars, errors = unbinned_energy_fit(peak_energy, peak, eres_pars, simplex=True)
    if np.isnan(pars[0]):
        print("first fit failed")
        pars, errors = unbinned_energy_fit(peak_energy, peak, eres_pars, simplex=False)
    pc_n = pars[0]
    pc_err = errors[0]
    sfs = []
    sf_errs = []
    for cut_val in cut_vals:
        sf, err, cut_pars, surv_pars = get_survival_fraction(
            peak_energy,
            peak_aoe,
            cut_val,
            peak,
            eres_pars,
            guess_pars_cut=None,
            guess_pars_surv=None,
        )
        sfs.append(sf)
        sf_errs.append(err)

    # return cut_vals, sfs, sf_errs
    ids = (
        (sf_errs < (1.5 * np.nanpercentile(sf_errs, 70)))
        & (~np.isnan(sf_errs))
        & (np.array(sfs) < 95)
    )
    fit = np.polynomial.polynomial.polyfit(
        cut_vals[ids], np.array(sfs)[ids], w=1 / np.array(sf_errs)[ids], deg=6
    )

    xs = np.arange(np.nanmin(cut_vals[ids]), np.nanmax(cut_vals[ids]), 0.01)
    p = np.polynomial.polynomial.polyval(xs, fit)
    cut_val = xs[np.argmin(np.abs(p - (100 * dep_acc)))]

    if display > 0:
        plt.figure()
        plt.errorbar(
            cut_vals[ids],
            np.array(sfs)[ids],
            yerr=np.array(sf_errs)[ids],
            linestyle=" ",
        )

        plt.plot(xs, p)
        plt.show()

    return cut_val


def get_sf(
    energy: np.array,
    aoe: np.array,
    peak: float,
    fit_width: tuple(int, int),
    aoe_cut_val: float,
    eres_pars: list,
    display: int = 0,
) -> tuple(np.array, np.array, np.array, float, float):
    """
    Calculates survival fraction for gamma lines using fitting method as in cut determination
    """

    # fwhm = np.sqrt(eres[0]+peak*eres[1])
    min_range = peak - fit_width[0]
    max_range = peak + fit_width[1]
    if peak == "1592.5":
        peak_energy = energy[(energy > min_range) & (energy < max_range)][:20000]
        peak_aoe = aoe[(energy > min_range) & (energy < max_range)][:20000]
    else:
        peak_energy = energy[(energy > min_range) & (energy < max_range)][:50000]
        peak_aoe = aoe[(energy > min_range) & (energy < max_range)][:50000]
    pars, errors = unbinned_energy_fit(peak_energy, peak, eres_pars, simplex=False)
    pc_n = pars[0]
    pc_err = errors[0]
    sfs = []
    sf_errs = []

    cut_vals = np.arange(-5, 5, 0.2)
    # cut_vals = np.append(cut_vals, aoe_cut_val)
    final_cut_vals = []
    for cut_val in cut_vals:
        try:
            sf, err, cut_pars, surv_pars = get_survival_fraction(
                peak_energy, peak_aoe, cut_val, peak, eres_pars
            )
            if np.isnan(cut_pars).all() == False and np.isnan(surv_pars).all() == False:
                guess_pars_cut = cut_pars
                guess_pars_surv = surv_pars
        except:
            sf = np.nan
            err = np.nan
        sfs.append(sf)
        sf_errs.append(err)
        final_cut_vals.append(cut_val)
    ids = (
        (sf_errs < (5 * np.nanpercentile(sf_errs, 50)))
        & (~np.isnan(sf_errs))
        & (np.array(sfs) < 100)
    )
    sf, sf_err, cut_pars, surv_pars = get_survival_fraction(
        peak_energy, peak_aoe, aoe_cut_val, peak, eres_pars
    )

    if display > 0:
        plt.figure()
        plt.errorbar(cut_vals, sfs, sf_errs)
        plt.show()

    return (
        np.array(final_cut_vals)[ids],
        np.array(sfs)[ids],
        np.array(sf_errs)[ids],
        sf,
        sf_err,
    )


def compton_sf(
    energy: np.array,
    aoe: np.array,
    cut: float,
    peak: float,
    eres: list[float, float],
    display: int = 1,
) -> tuple(float, np.array, list):
    """
    Determines survival fraction for compton continuum by basic counting
    """

    fwhm = np.sqrt(eres[0] + peak * eres[1])

    emin = peak - 2 * fwhm
    emax = peak + 2 * fwhm
    sfs = []
    aoe = aoe[(energy > emin) & (energy < emax)]
    cut_vals = np.arange(-5, 5, 0.1)
    for cut_val in cut_vals:
        sfs.append(100 * len(aoe[(aoe > cut_val)]) / len(aoe))
    sf = 100 * len(aoe[(aoe > cut)]) / len(aoe)
    return sf, cut_vals, sfs


def get_sf_no_sweep(
    energy: np.array,
    aoe: np.array,
    peak: float,
    fit_width: tuple(int, int),
    eres_pars: list,
    aoe_low_cut_val: float,
    aoe_high_cut_val: float = None,
    display: int = 1,
) -> tuple(float, float):
    """
    Calculates survival fraction for gamma line without sweeping through values
    """

    min_range = peak - fit_width[0]
    max_range = peak + fit_width[1]
    if peak == "1592.5":
        peak_energy = energy[(energy > min_range) & (energy < max_range)][:20000]
        peak_aoe = aoe[(energy > min_range) & (energy < max_range)][:20000]
    else:
        peak_energy = energy[(energy > min_range) & (energy < max_range)][:50000]
        peak_aoe = aoe[(energy > min_range) & (energy < max_range)][:50000]

    sf, sf_err, cut_pars, surv_pars = get_survival_fraction(
        peak_energy,
        peak_aoe,
        aoe_low_cut_val,
        peak,
        eres_pars,
        high_cut=aoe_high_cut_val,
    )
    return sf, sf_err


def compton_sf_no_sweep(
    energy: np.array,
    aoe: np.array,
    peak: float,
    eres: list[float, float],
    aoe_low_cut_val: float,
    aoe_high_cut_val: float = None,
    display: int = 1,
) -> float:
    """
    Calculates survival fraction for compton contiuum without sweeping through values
    """

    fwhm = np.sqrt(eres[0] + peak * eres[1])

    emin = peak - 2 * fwhm
    emax = peak + 2 * fwhm
    sfs = []
    aoe = aoe[(energy > emin) & (energy < emax)]
    cut_vals = np.arange(-5, 5, 0.1)
    if aoe_high_cut_val is None:
        sf = 100 * len(aoe[(aoe > aoe_low_cut_val)]) / len(aoe)
    else:
        sf = (
            100
            * len(aoe[(aoe > aoe_low_cut_val) & (aoe < aoe_high_cut_val)])
            / len(aoe)
        )
    return sf


def get_classifier(
    aoe: np.array,
    energy: np.array,
    mu_pars: list[float, float],
    sigma_pars: list[float, float],
) -> np.array:
    """
    Applies correction to A/E energy dependence
    """

    classifier = aoe / (mu_pars[0] * energy + mu_pars[1])
    classifier = (classifier - 1) / sigma_fit(energy, *sigma_pars)
    return classifier


def get_dt_guess(hist: np.array, bins: np.array, var: np.array) -> list:
    """
    Guess for fitting dt spectrum
    """

    mu, sigma, amp = pgh.get_gaussian_guess(hist, bins)
    i_0 = np.argmax(hist)
    bg = np.mean(hist[-10:])
    step = bg - np.mean(hist[:10])
    hstep = step / (bg + np.mean(hist[:10]))
    dx = np.diff(bins)[0]
    n_bins_range = int((3 * sigma) // dx)
    nsig_guess = np.sum(hist[i_0 - n_bins_range : i_0 + n_bins_range])
    nbkg_guess = np.sum(hist) - nsig_guess
    return [
        nsig_guess * np.diff(bins)[0],
        mu,
        sigma,
        nbkg_guess * np.diff(bins)[0],
        hstep,
        np.inf,
        np.inf,
        0,
    ]


def apply_dtcorr(aoe: np.array, dt: np.array, alpha: float) -> np.array:
    """Aligns dt regions"""
    return aoe * (1 + alpha * dt)


def drift_time_correction(
    aoe: np.array,
    energy: np.array,
    dt: np.array,
    display: int = 0,
    plot_dict: dict = {},
) -> tuple(np.array, float):
    """
    Calculates the correction needed to align the two drift time regions for ICPC detectors
    """
    hist, bins, var = pgh.get_hist(aoe[(energy > 1582) & (energy < 1602)], bins=500)
    bin_cs = (bins[1:] + bins[:-1]) / 2
    mu = bin_cs[np.argmax(hist)]
    aoe_range = [mu * 0.9, mu * 1.1]

    idxs = (
        (energy > 1582) & (energy < 1602) & (aoe > aoe_range[0]) & (aoe < aoe_range[1])
    )

    mask = (
        (idxs)
        & (dt < np.nanpercentile(dt[idxs], 55))
        & (dt > np.nanpercentile(dt[idxs], 1))
    )

    hist, bins, var = pgh.get_hist(
        dt[mask], dx=10, range=(np.nanmin(dt[mask]), np.nanmax(dt[mask]))
    )

    gpars = get_dt_guess(hist, bins, var)
    dt_pars, dt_errs, dt_cov = pgf.fit_binned(
        pgf.gauss_step_pdf,
        hist,
        bins,
        guess=gpars,
        fixed=[-3, -2, -1],
        cost_func="Least Squares",
    )

    aoe_mask = (
        (idxs) & (dt > dt_pars[1] - 2 * dt_pars[2]) & (dt < dt_pars[1] + 2 * dt_pars[2])
    )
    aoe_tmp = aoe[aoe_mask]
    aoe_pars, aoe_errs = unbinned_aoe_fit(aoe_tmp, display=display)

    mask2 = (
        (idxs)
        & (dt > np.nanpercentile(dt[idxs], 50))
        & (dt < np.nanpercentile(dt[idxs], 99))
    )
    hist2, bins2, var2 = pgh.get_hist(
        dt[mask2], dx=10, range=(np.nanmin(dt[mask2]), np.nanmax(dt[mask2]))
    )
    gpars2 = get_dt_guess(hist2, bins2, var2)

    dt_pars2, dt_errs2, dt_cov2 = pgf.fit_binned(
        pgf.gauss_step_pdf,
        hist2,
        bins2,
        guess=gpars2,
        fixed=[-3, -2, -1],
        cost_func="Least Squares",
    )

    aoe_mask2 = (
        (idxs)
        & (dt > dt_pars2[1] - 2 * dt_pars2[2])
        & (dt < dt_pars2[1] + 2 * dt_pars2[2])
    )
    aoe_tmp2 = aoe[aoe_mask2]
    aoe_pars2, aoe_errs2 = unbinned_aoe_fit(aoe_tmp2, display=display)

    alpha = (aoe_pars[2] - aoe_pars2[2]) / (
        dt_pars2[1] * aoe_pars2[2] - dt_pars[1] * aoe_pars[2]
    )
    aoe_corrected = apply_dtcorr(aoe, dt, alpha)

    if display > 0:
        dt_fig = plt.figure()
        plt.subplot(3, 2, 1)
        plt.step(pgh.get_bin_centers(bins), hist, label="Data")
        plt.plot(
            pgh.get_bin_centers(bins),
            pgf.gauss_step_pdf(pgh.get_bin_centers(bins), *gpars),
            label="Guess",
        )
        plt.plot(
            pgh.get_bin_centers(bins),
            pgf.gauss_step_pdf(pgh.get_bin_centers(bins), *dt_pars),
            label="Fit",
        )
        plt.xlabel("Drift Time (ns)")
        plt.ylabel("Counts")
        plt.legend(loc="upper left")

        plt.subplot(3, 2, 2)
        plt.step(pgh.get_bin_centers(bins2), hist2, label="Data")
        plt.plot(
            pgh.get_bin_centers(bins2),
            pgf.gauss_step_pdf(pgh.get_bin_centers(bins2), *gpars2),
            label="Guess",
        )
        plt.plot(
            pgh.get_bin_centers(bins2),
            pgf.gauss_step_pdf(pgh.get_bin_centers(bins2), *dt_pars2),
            label="Fit",
        )
        plt.xlabel("Drift Time (ns)")
        plt.ylabel("Counts")
        plt.legend(loc="upper left")

        plt.subplot(3, 2, 3)
        xs = np.linspace(aoe_pars[-3], aoe_pars[-2], 1000)
        counts, aoe_bins, bars = plt.hist(
            aoe[(aoe < aoe_pars[-2]) & (aoe > aoe_pars[-3]) & aoe_mask],
            bins=400,
            histtype="step",
            label="Data",
        )
        dx = np.diff(aoe_bins)
        plt.plot(xs, PDF_AoE(xs, *aoe_pars)[1] * dx[0], label="Full fit")
        # plt.yscale('log')
        n_events, sig, bkg = PDF_AoE(xs, *aoe_pars[:-1], True)
        plt.plot(xs, sig * dx[0], label="Peak fit")
        plt.plot(xs, bkg * dx[0], label="Bkg fit")
        plt.legend(loc="upper left")
        plt.xlabel("A/E")
        plt.ylabel("Counts")

        plt.subplot(3, 2, 4)
        xs = np.linspace(aoe_pars2[-3], aoe_pars2[-2], 1000)
        counts, aoe_bins2, bars = plt.hist(
            aoe[(aoe < aoe_pars2[-2]) & (aoe > aoe_pars2[-3]) & aoe_mask2],
            bins=400,
            histtype="step",
            label="Data",
        )
        dx = np.diff(aoe_bins2)
        plt.plot(xs, PDF_AoE(xs, *aoe_pars2)[1] * dx[0], label="Full fit")
        # plt.yscale('log')
        n_events, sig, bkg = PDF_AoE(xs, *aoe_pars2[:-1], True)
        plt.plot(xs, sig * dx[0], label="Peak fit")
        plt.plot(xs, bkg * dx[0], label="Bkg fit")
        plt.legend(loc="upper left")
        plt.xlabel("A/E")
        plt.ylabel("Counts")

        plt.subplot(3, 2, 5)
        counts, bins, bars = plt.hist(
            aoe[idxs], bins=200, histtype="step", label="Uncorrected"
        )
        plt.hist(aoe_corrected[idxs], bins=bins, histtype="step", label="Corrected")
        plt.xlabel("A/E")
        plt.ylabel("Counts")
        plt.legend(loc="upper left")
        plt.tight_layout()
        plt.xlim([aoe_pars[-3], aoe_pars[-2] * 1.01])
        plot_dict["dt_corr"] = dt_fig
        if display > 1:
            plt.show()
        else:
            plt.close()
        return aoe_corrected, alpha, plot_dict
    else:
        return aoe_corrected, alpha


def cal_aoe(
    files: list,
    lh5_path,
    cal_dict: dict,
    energy_param: str,
    cal_energy_param: str,
    eres_pars: list,
    cut_field: str = "is_valid_cal",
    dt_corr: bool = False,
    aoe_high_cut: int = 4,
    display: int = 0,
) -> tuple(dict, dict):
    """
    Main function for running the a/e correction and cut determination.
    """

    aoe_uncorr, energy, dt = load_aoe(
        files, lh5_path, cal_dict, energy_param, cal_energy_param, cut_field=cut_field
    )

    if dt_corr == True:
        aoe, alpha = drift_time_correction(aoe_uncorr, energy, dt)
    else:
        aoe = aoe_uncorr

    try:
        aoe_tmp = aoe[(energy > 1000) & (energy < 1300) & (aoe > 0)]  # [:20000]
        bulk_pars, bulk_errs = unbinned_aoe_fit(aoe_tmp, display=0)
    except:
        bulk_pars = np.full(8, np.nan)
        log.error("1000-1300 keV mean determination failed")

    try:
        log.info("Starting A/E correction")
        mu_pars, sigma_pars, results_dict = AoEcorrection(energy, aoe, eres_pars)
        log.info("Finished A/E correction")

        classifier = get_classifier(aoe, energy, mu_pars, sigma_pars)

    except:
        log.error("A/E calibration failed")
        mean_pars = np.full(2, np.nan)
        sigma_pars = np.full(2, np.nan)

    try:
        cut = get_aoe_cut_fit(
            energy, classifier, 1592, (40, 20), 0.9, eres_pars, display=0
        )
    except:
        log.error("A/E cut determination failed")
        cut = np.nan

    cal_dict.update(
        {
            "AoE_Corrected": {
                "expression": f"(((A_max/{energy_param})/(a*{cal_energy_param} +b))-1)",
                "parameters": {"a": mu_pars[0], "b": mu_pars[1]},
            },
            "AoE_Classifier": {
                "expression": f"AoE_Corrected/(sqrt(c+(d/{cal_energy_param})**2)/(a*{cal_energy_param} +b))",
                "parameters": {
                    "a": mu_pars[0],
                    "b": mu_pars[1],
                    "c": sigma_pars[0],
                    "d": sigma_pars[1],
                },
            },
            "AoE_Low_Cut": {
                "expression": "AoE_Classifier>a",
                "parameters": {"a": cut},
            },
            "AoE_Double_Sided_Cut": {
                "expression": "(b>AoE_Classifier)&(AoE_Classifier>a)",
                "parameters": {"a": cut, "b": aoe_high_cut},
            },
        }
    )

    try:
        log.info("  Compute low side survival fractions: ")

        peaks_of_interest = [1592.5, 1620.5, 2039, 2103.53, 2614.50]
        sf = np.zeros(len(peaks_of_interest))
        sferr = np.zeros(len(peaks_of_interest))
        fit_widths = [(40, 25), (25, 40), (0, 0), (25, 40), (50, 50)]
        full_sfs = []
        full_sf_errs = []
        full_cut_vals = []

        for i, peak in enumerate(peaks_of_interest):
            if peak == 2039:
                sf[i], cut_vals, sfs = compton_sf(
                    energy, classifier, cut, peak, eres_pars
                )
                sferr[i] = 0

                full_cut_vals.append(cut_vals)
                full_sfs.append(sfs)
                full_sf_errs.append(None)
            else:
                cut_vals, sfs, sf_errs, sf[i], sferr[i] = get_sf(
                    energy, classifier, peak, fit_widths[i], cut, eres_pars
                )
                full_cut_vals.append(cut_vals)
                full_sfs.append(sfs)
                full_sf_errs.append(sf_errs)

            log.info(f"{peak}keV: {sf[i]:2.1f} +/- {sferr[i]:2.1f} %")

        sf_2side = np.zeros(len(peaks_of_interest))
        sferr_2side = np.zeros(len(peaks_of_interest))
        log.info("Calculating 2 sided cut sfs")
        for i, peak in enumerate(peaks_of_interest):
            if peak == 2039:
                sf_2side[i] = compton_sf_no_sweep(
                    energy,
                    classifier,
                    peak,
                    eres_pars,
                    cut,
                    aoe_high_cut_val=aoe_high_cut,
                )
                sferr_2side[i] = 0
            else:
                sf_2side[i], sferr_2side[i] = get_sf_no_sweep(
                    energy,
                    classifier,
                    peak,
                    fit_widths[i],
                    eres_pars,
                    cut,
                    aoe_high_cut_val=aoe_high_cut,
                )

            log.info(f"{peak}keV: {sf[i]:2.1f} +/- {sferr[i]:2.1f} %")

        def convert_sfs_to_dict(peaks_of_interest, sfs, sf_errs):
            out_dict = {}
            for i, peak in enumerate(peaks_of_interest):
                out_dict[str(peak)] = {
                    "sf": f"{sfs[i]:2f}",
                    "sf_err": f"{sf_errs[i]:2f}",
                }
            return out_dict

        out_dict = {
            "correction_fit_results": results_dict,
            "A/E_Energy_param": energy_param,
            "Cal_energy_param": cal_energy_param,
            "dt_param": "dt_eff",
            "rt_correction": dt_corr,
            "1000-1300keV_mean": bulk_pars[2],
            "Mean_pars": list(mu_pars),
            "Sigma_pars": list(sigma_pars),
            "Low_cut": cut,
            "High_cut": aoe_high_cut,
            "Low_side_sfs": convert_sfs_to_dict(peaks_of_interest, sf, sferr),
            "2_side_sfs": convert_sfs_to_dict(peaks_of_interest, sf_2side, sferr_2side),
        }
        log.info("Done")
        log.info(f"Results are {out_dict}")

        if display > 0:
            plot_dict = {}

            plt.rcParams["figure.figsize"] = (12, 8)
            plt.rcParams["font.size"] = 16

            fig1 = plt.figure()
            plt.subplot(3, 2, 1)
            plot_dt_dep(aoe, energy, dt, [1582, 1602], f"Tl DEP")
            plt.subplot(3, 2, 2)
            plot_dt_dep(aoe, energy, dt, [1510, 1630], f"Bi FEP")
            plt.subplot(3, 2, 3)
            plot_dt_dep(aoe, energy, dt, [2030, 2050], "Qbb")
            plt.subplot(3, 2, 4)
            plot_dt_dep(aoe, energy, dt, [2080, 2120], f"Tl SEP")
            plt.subplot(3, 2, 5)
            plot_dt_dep(aoe, energy, dt, [2584, 2638], f"Tl FEP")
            plt.tight_layout()
            plot_dict["dt_deps"] = fig1
            if display > 1:
                plt.show()
            else:
                plt.close()

            fig2 = plt.figure()
            plot_compt_bands_overlayed(aoe, energy, [950, 1250, 1460, 1660, 1860, 2060])
            plt.ylabel("Counts")
            plt.xlabel("Raw A/E")
            plt.title(f"Compton Bands before Correction")
            plt.legend(loc="upper left")
            plot_dict["compt_bands_nocorr"] = fig2
            if display > 1:
                plt.show()
            else:
                plt.close()

            if dt_corr == True:
                _, plot_dict = drift_time_correction(
                    aoe_uncorr, energy, dt, display=display, plot_dict=plot_dict
                )

            mu_pars, sigma_pars, results_dict, plot_dict = AoEcorrection(
                energy, aoe, eres_pars, plot_dict, display=display
            )

            fig3 = plt.figure()
            plot_compt_bands_overlayed(
                classifier, energy, [950, 1250, 1460, 1660, 1860, 2060], [-5, 5]
            )
            plt.ylabel("Counts")
            plt.xlabel("Corrected A/E")
            plt.title(f"Compton Bands after Correction")
            plt.legend(loc="upper left")
            plot_dict["compt_bands_corr"] = fig3
            if display > 1:
                plt.show()
            else:
                plt.close()

            fig4 = plt.figure()
            plt.vlines(cut, 0, 100, label=f"Cut Value: {cut:1.2f}", color="black")

            for i, peak in enumerate(peaks_of_interest):
                if peak == 2039:
                    plt.plot(
                        full_cut_vals[i],
                        full_sfs[i],
                        label=f"{get_peak_label(peak)} {peak} keV: {sf[i]:2.1f} +/- {sferr[i]:2.1f} %",
                    )
                else:
                    plt.errorbar(
                        full_cut_vals[i],
                        full_sfs[i],
                        yerr=full_sf_errs[i],
                        label=f"{get_peak_label(peak)} {peak} keV: {sf[i]:2.1f} +/- {sferr[i]:2.1f} %",
                    )

            handles, labels = plt.gca().get_legend_handles_labels()
            order = [1, 2, 3, 0, 4, 5]
            plt.legend(
                [handles[idx] for idx in order],
                [labels[idx] for idx in order],
                loc="upper right",
            )
            plt.xlabel("Cut Value")
            plt.ylabel("Survival Fraction %")
            plot_dict["surv_fracs"] = fig4
            if display > 1:
                plt.show()
            else:
                plt.close()

            fig5, ax = plt.subplots()
            bins = np.linspace(1000, 3000, 2000)
            ax.hist(energy, bins=bins, histtype="step", label="Before PSD")
            ax.hist(
                energy[classifier > cut],
                bins=bins,
                histtype="step",
                label="Low side PSD cut",
            )
            ax.hist(
                energy[(classifier > cut) & (classifier < 4)],
                bins=bins,
                histtype="step",
                label="Double sided PSD cut",
            )
            ax.hist(
                energy[(classifier < cut) | (classifier > 4)],
                bins=bins,
                histtype="step",
                label="Rejected by PSD cut",
            )

            axins = ax.inset_axes([0.25, 0.07, 0.4, 0.3])
            bins = np.linspace(1580, 1640, 200)
            axins.hist(energy, bins=bins, histtype="step")
            axins.hist(energy[classifier > cut], bins=bins, histtype="step")
            axins.hist(
                energy[(classifier > cut) & (classifier < aoe_high_cut)],
                bins=bins,
                histtype="step",
            )
            axins.hist(
                energy[(classifier < cut) | (classifier > aoe_high_cut)],
                bins=bins,
                histtype="step",
            )
            ax.set_xlim([1000, 3000])
            ax.set_yscale("log")
            plt.xlabel("Energy (keV)")
            plt.ylabel("Counts")
            plt.legend(loc="upper left")
            plot_dict["PSD_spectrum"] = fig5
            if display > 1:
                plt.show()
            else:
                plt.close()

            fig6 = plt.figure()
            bins = np.linspace(1000, 3000, 1000)
            counts_pass, bins_pass, _ = pgh.get_hist(
                energy[(classifier > cut) & (classifier < aoe_high_cut)], bins=bins
            )
            counts, bins, _ = pgh.get_hist(energy, bins=bins)
            survival_fracs = counts_pass / (counts)

            plt.step(pgh.get_bin_centers(bins_pass), survival_fracs)
            plt.xlabel("Energy (keV)")
            plt.ylabel("Survival Fraction")
            plt.ylim([0, 1])
            plot_dict["psd_sf"] = fig6
            if display > 1:
                plt.show()
            else:
                plt.close()

            return cal_dict, out_dict, plot_dict
        else:
            return cal_dict, out_dict

    except:
        log.error("A/E Survival fraction determination failed")
        out_dict = {
            "correction_fit_results": results_dict,
            "A/E_Energy_param": "cuspEmax",
            "Cal_energy_param": "cuspEmax_ctc",
            "dt_param": "dt_eff",
            "rt_correction": False,
            "1000-1300keV_mean": bulk_pars[2],
            "Mean_pars": list(mu_pars),
            "Sigma_pars": list(sigma_pars),
            "Low_cut": cut,
            "High_cut": aoe_high_cut,
        }
        if display > 0:
            plot_dict = {}
            return cal_dict, out_dict, plot_dict
        else:
            return cal_dict, out_dict



from __future__ import annotations

from math import exp, log

import numpy as np
from numba import guvectorize

from pygama.dsp.utils import numba_defaults_kwargs as nb_kwargs


@guvectorize(
    [
        "void(float32[:], float32, float32, float32, float32, float32[:])",
        "void(float64[:], float64, float64,float64,float64,float64[:])",
    ],
    "(n),(),(),(), ()->(n)",
    **nb_kwargs,
)
def inject_sig_pulse(
    wf_in: np.ndarray, t0: int, rt: float, a: float, decay: float, wf_out: np.ndarray
) -> None:
    r"""Inject sigmoid pulse into existing waveform to simulate pileup.

    .. math::
        s(t) = \frac{A}{1 + \exp[-4 \log(99) (t - t_0 - t_r/2) / t_r]}
                e^{-(t-t_0)/\tau}

    Parameters
    ----------
    wf_in
        the input waveform.
    t0
        the position :math:`t_0` of the injected waveform.
    rt
        the rise time :math:`t_r` of the injected waveform.
    a
        the amplitude :math:`A` of the injected waveform.
    decay
        the decay parameter :math:`\tau` of the injected waveform.
    wf_out
        the output waveform.
    """

    wf_out[:] = np.nan

    if (
        np.isnan(wf_in).any()
        or np.isnan(rt)
        or np.isnan(t0)
        or np.isnan(a)
        or np.isnan(decay)
    ):
        return

    rise = 4 * log(99) / rt

    wf_out[:] = wf_in[:]
    for t in range(len(wf_out)):
        wf_out[t] = wf_out[t] + a / (1 + exp(-rise * (t - (t0 + rt / 2)))) * exp(
            -(1 / decay) * (t - t0)
        )


@guvectorize(
    [
        "void(float32[:], float32, float32, float32, float32, float32[:])",
        "void(float64[:], float64, float64,float64,float64,float64[:])",
    ],
    "(n),(),(),(), ()->(n)",
    **nb_kwargs,
)
def inject_exp_pulse(
    wf_in: np.ndarray, t0: int, rt: float, a: float, decay: float, wf_out: np.ndarray
) -> None:
    """Inject exponential pulse into existing waveform to simulate pileup.

    Parameters
    ----------
    wf_in
        the input waveform.
    t0
        the position of the injected waveform.
    rt
        the rise time of the injected waveform.
    a
        the amplitude of the injected waveform.
    decay
        the exponential decay constant of the injected waveform.
    wf_out
        the output waveform.
    """

    wf_out[:] = np.nan

    if (
        np.isnan(wf_in).any()
        or np.isnan(rt)
        or np.isnan(t0)
        or np.isnan(a)
        or np.isnan(decay)
    ):
        return

    wf_out[:] = wf_in[:]
    for t in range(len(wf_out)):
        if t <= t0 and t <= t0 + rt:
            wf_out[t] += a * exp((t - t0 - rt) / (rt)) * exp(-(1 / decay) * (t - t0))
        elif t > t0 + rt:
            wf_out[t] += a * exp(-(1 / decay) * (t - t0))



import logging
import multiprocessing as mp
from collections import namedtuple
from multiprocessing import get_context
from pprint import pprint

import numpy as np

from pygama.dsp import build_processing_chain

log = logging.getLogger(__name__)


def run_one_dsp(
    tb_data, dsp_config, db_dict=None, fom_function=None, verbosity=0, fom_kwargs=None
):
    """
    run one iteration of DSP on tb_data

    Optionally returns a value for optimization

    Parameters
    ----------
    tb_data : lh5 Table
        An input table of lh5 data. Typically a selection is made prior to
        sending tb_data to this function: optimization typically doesn't have to
        run over all data
    dsp_config : dict
        Specifies the DSP to be performed for this iteration (see
        build_processing_chain()) and the list of output variables to appear in
        the output table
    db_dict : dict (optional)
        DSP parameters database. See build_processing_chain for formatting info
    fom_function : function or None (optional)
        When given the output lh5 table of this DSP iteration, the
        fom_function must return a scalar figure-of-merit value upon which the
        optimization will be based. Should accept verbosity as a second argument
    verbosity : int (optional)
        verbosity for the processing chain and fom_function calls
    fom_kwargs
        any keyword arguments to pass to the fom

    Returns
    -------
    figure_of_merit : float
        If fom_function is not None, returns figure-of-merit value for the DSP iteration
    tb_out : lh5 Table
        If fom_function is None, returns the output lh5 table for the DSP iteration
    """

    pc, lh5_col_names, tb_out = build_processing_chain(
        tb_data, dsp_config, db_dict=db_dict
    )
    pc.execute()
    if fom_function is not None:
        if fom_kwargs is not None:
            return fom_function(tb_out, verbosity, fom_kwargs)
        else:
            return fom_function(tb_out, verbosity)
    else:
        return tb_out


ParGridDimension = namedtuple("ParGridDimension", "name parameter value_strs")


class ParGrid:
    """Parameter Grid class
    Each ParGrid entry corresponds to a dsp parameter to be varied.
    The ntuples must follow the pattern:
    ( name parameter value_strs) : ( str, str, list of str)
    where name and parameter are the same as 'db.name.parameter' in the processing chain,
    value_strs is the array of strings to set the argument to.
    """

    def __init__(self):
        self.dims = []

    def add_dimension(self, name, parameter, value_strs):
        self.dims.append(ParGridDimension(name, parameter, value_strs))

    def get_n_dimensions(self):
        return len(self.dims)

    def get_n_points_of_dim(self, i):
        return len(self.dims[i].value_strs)

    def get_shape(self):
        shape = ()
        for i in range(self.get_n_dimensions()):
            shape += (self.get_n_points_of_dim(i),)
        return shape

    def get_n_grid_points(self):
        return np.prod(self.get_shape())

    def get_par_meshgrid(self, copy=False, sparse=False):
        """return a meshgrid of parameter values
        Always uses Matrix indexing (natural for par grid) so that
        mg[i1][i2][...] corresponds to index order in self.dims
        Note copy is False by default as opposed to numpy default of True
        """
        axes = []
        for i in range(self.get_n_dimensions()):
            axes.append(self.dims[i].values_strs)
        return np.meshgrid(*axes, copy, sparse, indexing="ij")

    def get_zero_indices(self):
        return np.zeros(self.get_n_dimensions(), dtype=np.uint32)

    def iterate_indices(self, indices):
        """iterate given indices [i1, i2, ...] by one.
        For easier iteration. The convention here is arbitrary, but its the
        order the arrays would be traversed in a series of nested for loops in
        the order appearin in dims (first dimension is first for loop, etc):
        Return False when the grid runs out of indices. Otherwise returns True.
        """
        for iD in reversed(range(self.get_n_dimensions())):
            indices[iD] += 1
            if indices[iD] < self.get_n_points_of_dim(iD):
                return True
            indices[iD] = 0
        return False

    # def check_indices(self, indices):
    #    for iD in reversed(range(self.get_n_dimensions())):
    #        if indices[iD] < self.get_n_points_of_dim(iD): return True
    #        indices[iD] = 0
    #    return False

    def get_data(self, i_dim, i_par):
        name = self.dims[i_dim].name
        parameter = self.dims[i_dim].parameter
        value_str = self.dims[i_dim].value_strs[i_par]
        return name, parameter, value_str

    def print_data(self, indices):
        print_string = f"Grid point at indices {indices}:"
        for i_dim, i_par in enumerate(indices):
            name, parameter, value_str = self.get_data(i_dim, i_par)
            print_string += f"\n {name}.{parameter} = {value_str}"
        return print_string

    def set_dsp_pars(self, db_dict, indices):
        if db_dict is None:
            db_dict = {}
        for i_dim, i_par in enumerate(indices):
            name, parameter, value_str = self.get_data(i_dim, i_par)
            if name not in db_dict.keys():
                db_dict[name] = {parameter: value_str}
            else:
                db_dict[name][parameter] = value_str
        return db_dict


def run_grid(
    tb_data, dsp_config, grid, fom_function, db_dict=None, verbosity=1, **fom_kwargs
):
    """Extract a table of optimization values for a grid of DSP parameters
    The grid argument defines a list of parameters and values over which to run
    the DSP defined in dsp_config on tb_data. At each point, a scalar
    figure-of-merit is extracted.

    Returns a N-dimensional ndarray of figure-of-merit values, where the array
    axes are in the order they appear in grid.

    Parameters
    ----------
    tb_data : lh5 Table
        An input table of lh5 data. Typically a selection is made prior to
        sending tb_data to this function: optimization typically doesn't have to
        run over all data
    dsp_config : dict
        Specifies the DSP to be performed (see build_processing_chain()) and the
        list of output variables to appear in the output table for each grid point
    grid : ParGrid
        See ParGrid class for format
    fom_function : function
        When given the output lh5 table of this DSP iteration, the fom_function
        must return a scalar figure-of-merit. Should accept verbosity as a
        second keyword argument
    db_dict : dict (optional)
        DSP parameters database. See build_processing_chain for formatting info
    verbosity : int (optional)
        verbosity for the processing chain and fom_function calls
    **fom_kwargs
        Any keyword arguments for fom_function

    Returns
    -------
    grid_values : ndarray of floats
        An N-dimensional numpy ndarray whose Mth axis corresponds to the Mth row
        of the grid argument
    """

    grid_values = np.ndarray(shape=grid.get_shape(), dtype="O")
    iii = grid.get_zero_indices()
    log.info("starting grid calculations...")
    while True:
        db_dict = grid.set_dsp_pars(db_dict, iii)
        if verbosity > 1:
            pprint(dsp_config)
        log.debug(grid.print_data(iii))
        grid_values[tuple(iii)] = run_one_dsp(
            tb_data,
            dsp_config,
            db_dict=db_dict,
            fom_function=fom_function,
            verbosity=verbosity,
            fom_kwargs=fom_kwargs,
        )
        log.debug("value:", grid_values[tuple(iii)])
        if not grid.iterate_indices(iii):
            break
    return grid_values


def run_grid_point(
    tb_data,
    dsp_config,
    grids,
    fom_function,
    iii,
    db_dict=None,
    verbosity=1,
    fom_kwargs=None,
):
    """
    Runs a single grid point for the index specified
    """
    if not isinstance(grids, list):
        grids = [grids]
    for index, grid in zip(iii, grids):
        db_dict = grid.set_dsp_pars(db_dict, index)

    log.debug(db_dict)
    for i, grid in enumerate(grids):
        log.debug(grid.print_data(iii[i]))
    tb_out = run_one_dsp(tb_data, dsp_config, db_dict=db_dict, verbosity=verbosity)
    res = np.ndarray(shape=len(grids), dtype="O")
    if fom_function:
        for i in range(len(grids)):
            if fom_kwargs[i] is not None:
                if len(fom_function) > 1:
                    res[i] = fom_function[i](tb_out, verbosity, fom_kwargs[i])
                else:
                    res[i] = fom_function[0](tb_out, verbosity, fom_kwargs[i])
            else:
                if len(fom_function) > 1:
                    res[i] = fom_function[i](tb_out, verbosity)
                else:
                    res[i] = fom_function[0](tb_out, verbosity)
        log.debug("value:", res)
        out = {"indexes": [tuple(ii) for ii in iii], "results": res}

    else:
        out = {"indexes": [tuple(ii) for ii in iii], "results": tb_out}
    return out


def get_grid_points(grid):
    """
    Generates a list of the indices of all possible grid points
    """
    out = []
    iii = [gri.get_zero_indices() for gri in grid]
    complete = np.full(len(grid), False)
    while True:
        out.append([tuple(ii) for ii in iii])

        for i, gri in enumerate(zip(iii, grid)):
            if not gri[1].iterate_indices(gri[0]):
                log.info(f"{i} grid end")
                iii[i] = gri[1].get_zero_indices()
                complete[i] = True
        if all(complete):
            break
    return out


def run_grid_multiprocess_parallel(
    tb_data,
    dsp_config,
    grid,
    fom_function,
    db_dict=None,
    verbosity=1,
    processes=5,
    fom_kwargs=None,
):
    """
    run one iteration of DSP on tb_data with multiprocessing, can handle
    multiple grids if they are the same dimensions

    Optionally returns a value for optimization

    Parameters
    ----------
    tb_data : lh5 Table
        An input table of lh5 data. Typically a selection is made prior to
        sending tb_data to this function: optimization typically doesn't have to
        run over all data
    dsp_config : dict
        Specifies the DSP to be performed for this iteration (see
        build_processing_chain()) and the list of output variables to appear in
        the output table
    grid : pargrid, list of pargrids
        Grids to run optimization on
    db_dict : dict (optional)
        DSP parameters database. See build_processing_chain for formatting info
    fom_function : function or None (optional)
        When given the output lh5 table of this DSP iteration, the
        fom_function must return a scalar figure-of-merit value upon which the
        optimization will be based. Should accept verbosity as a second argument.
        If multiple grids provided can either pass one fom to have it run for each grid
        or a list of fom to run different fom on each grid.
    verbosity : int (optional)
        verbosity for the processing chain and fom_function calls
    processes : int
        DOCME
    fom_kwargs
        any keyword arguments to pass to the fom,
        if multiple grids given will need to be a list of the fom_kwargs for each grid

    Returns
    -------
    figure_of_merit : float
        If fom_function is not None, returns figure-of-merit value for the DSP iteration
    tb_out : lh5 Table
        If fom_function is None, returns the output lh5 table for the DSP iteration
    """

    if not isinstance(grid, list):
        grid = [grid]
    if not isinstance(fom_function, list) and fom_function is not None:
        fom_function = [fom_function]
    if not isinstance(fom_kwargs, list):
        fom_kwargs = [fom_kwargs for gri in grid]
    grid_values = []
    shapes = [gri.get_shape() for gri in grid]
    if fom_function is not None:
        for i in range(len(grid)):
            grid_values.append(np.ndarray(shape=shapes[i], dtype="O"))
    else:
        grid_lengths = np.array([gri.get_n_grid_points() for gri in grid])
        grid_values.append(np.ndarray(shape=shapes[np.argmax(grid_lengths)], dtype="O"))
    grid_list = get_grid_points(grid)
    pool = mp.Pool(processes=processes)
    results = [
        pool.apply_async(
            run_grid_point,
            args=(
                tb_data,
                dsp_config,
                grid,
                fom_function,
                np.asarray(gl),
                db_dict,
                verbosity,
                fom_kwargs,
            ),
        )
        for gl in grid_list
    ]

    for result in results:
        res = result.get()
        indexes = res["indexes"]
        if fom_function is not None:
            for i in range(len(grid)):
                index = indexes[i]
                if grid_values[i][index] is None:
                    grid_values[i][index] = res["results"][i]
        else:
            grid_values[0][indexes[0]] = {f"{indexes[0]}": res["results"]}

    pool.close()
    pool.join()
    return grid_values



import numpy as np
import pandas as pd
import pytest

import pygama.lgdo as lgdo
from pygama.lgdo.table import Table


def test_init():
    tbl = Table()
    assert tbl.size == 1024
    assert tbl.loc == 0

    tbl = Table(size=10)
    assert tbl.size == 10

    col_dict = {
        "a": lgdo.Array(nda=np.array([1, 2, 3, 4])),
        "b": lgdo.Array(nda=np.array([5, 6, 7, 8])),
    }

    tbl = Table(col_dict=col_dict)
    assert tbl.size == 4

    tbl = Table(size=3, col_dict=col_dict)
    assert tbl.size == 3


def test_datatype_name():
    tbl = Table()
    assert tbl.datatype_name() == "table"


def test_push_row():
    tbl = Table()
    tbl.push_row()
    assert tbl.loc == 1


def test_is_full():
    tbl = Table(size=2)
    tbl.push_row()
    assert tbl.is_full() is False
    tbl.push_row()
    assert tbl.is_full() is True


def test_clear():
    tbl = Table()
    tbl.push_row()
    tbl.clear()
    assert tbl.loc == 0


def test_add_field():
    tbl = Table()
    tbl.add_field("a", lgdo.Array(np.array([1, 2, 3])), use_obj_size=True)
    assert tbl.size == 3

    with pytest.raises(TypeError):
        tbl.add_field("s", lgdo.Scalar(value=69))


def test_add_column():
    tbl = Table()
    tbl.add_column("a", lgdo.Array(np.array([1, 2, 3])), use_obj_size=True)
    assert tbl.size == 3


def test_join():
    tbl1 = Table(size=3)
    tbl1.add_field("a", lgdo.FixedSizeArray(np.array([1, 2, 3])))
    tbl1.add_field("b", lgdo.Array(np.array([1, 2, 3])))
    assert list(tbl1.keys()) == ["a", "b"]

    tbl2 = Table(size=3)
    tbl2.add_field("c", lgdo.Array(np.array([4, 5, 6])))
    tbl2.add_field("d", lgdo.Array(np.array([9, 9, 10])))

    tbl1.join(tbl2)
    assert list(tbl1.keys()) == ["a", "b", "c", "d"]

    tbl2.join(tbl1, cols=("a"))
    assert list(tbl2.keys()) == ["c", "d", "a"]


def test_get_dataframe():
    tbl = Table(4)
    tbl.add_column("a", lgdo.Array(np.array([1, 2, 3])))
    tbl.add_column("b", lgdo.Array(np.array([[0, 1, 2], [3, 4, 5], [6, 7, 8]])))
    tbl.add_column(
        "c",
        lgdo.VectorOfVectors(
            flattened_data=lgdo.Array(np.array([0, 1, 2, 3, 4, 5, 6])),
            cumulative_length=lgdo.Array(np.array([3, 4, 7])),
        ),
    )
    tbl.add_column(
        "d",
        lgdo.Table(
            col_dict={
                "a": lgdo.Array(np.array([2, 4, 6, 8])),
                "b": lgdo.Array(np.array([[1, 1], [2, 4], [3, 9], [4, 16]])),
            }
        ),
    )
    df = tbl.get_dataframe()
    assert isinstance(df, pd.DataFrame)
    assert list(df.keys()) == ["a", "b", "c", "d_a", "d_b"]


def test_remove_column():
    col_dict = {
        "a": lgdo.Array(nda=np.array([1, 2, 3, 4])),
        "b": lgdo.Array(nda=np.array([5, 6, 7, 8])),
        "c": lgdo.Array(nda=np.array([9, 10, 11, 12])),
    }

    tbl = Table(col_dict=col_dict)

    tbl.remove_column("a")
    assert list(tbl.keys()) == ["b", "c"]

    tbl.remove_column("c")
    assert list(tbl.keys()) == ["b"]



from __future__ import annotations


class DSPError(Exception):
    """Base class for signal processors."""

    pass


class DSPFatal(DSPError):
    """Fatal error thrown by DSP processors that halts production.

    Attributes
    ----------
    wf_range: range
        range of wf indices. This will be set after the exception is caught,
        and appended to the error message
    processor: str
        string of processor and arguments. This will be set after the exception
        is caught, and appended to the error message
    """

    def __init__(self, *args) -> None:
        super().__init__(*args)
        self.wf_range = None
        self.processor = None

    def __str__(self) -> str:
        suffix = ""
        if self.wf_range:
            suffix += "\nThrown while processing entries " + str(self.wf_range)
        if self.processor:
            suffix += "\nThrown by " + self.processor
        return super().__str__() + suffix


class ProcessingChainError(DSPError):
    """Error thrown when there is a problem setting up a processing chain."""

    pass



import numpy as np

import pygama.lgdo as lgdo


def test_datatype_name():
    aoesa = lgdo.ArrayOfEqualSizedArrays()
    assert aoesa.datatype_name() == "array_of_equalsized_arrays"


def test_form_datatype():
    aoesa = lgdo.ArrayOfEqualSizedArrays(dims=(2, 3))
    assert aoesa.form_datatype() == "array_of_equalsized_arrays<2,3>{real}"


def test_init():
    attrs = {"attr1": 1}
    aoesa = lgdo.ArrayOfEqualSizedArrays(
        dims=(2, 3), dtype=np.float32, fill_val=42, attrs=attrs
    )
    assert aoesa.dims == (2, 3)
    assert (aoesa.nda == np.full((2, 3), 42, np.float32)).all()
    assert aoesa.attrs == attrs | {"datatype": "array_of_equalsized_arrays<2,3>{real}"}


def test_to_vov():
    aoesa = lgdo.ArrayOfEqualSizedArrays(
        nda=np.array([[53, 91, 66, 58, 8], [78, 57, 66, 88, 73], [85, 99, 86, 68, 53]])
    )
    vov = aoesa.to_vov()
    assert isinstance(vov, lgdo.VectorOfVectors)
    assert np.array_equal(vov[0], [53, 91, 66, 58, 8])
    assert np.array_equal(vov[1], [78, 57, 66, 88, 73])
    assert np.array_equal(vov[2], [85, 99, 86, 68, 53])

    vov = aoesa.to_vov([2, 5, 6])
    assert np.array_equal(vov[0], [53, 91])
    assert np.array_equal(vov[1], [78, 57, 66])
    assert np.array_equal(vov[2], [85])



import os
from pathlib import Path

import numpy as np

import pygama.lgdo.lh5_store as store
from pygama.dsp import build_dsp

config_dir = Path(__file__).parent / "configs"
dsp_file = "/tmp/LDQTA_r117_20200110T105115Z_cal_geds__numpy_test_dsp.lh5"


def test_list_parisng(lgnd_test_data):
    dsp_config = {
        "outputs": ["wf_out"],
        "processors": {
            "wf_out": {
                "function": "add",
                "module": "numpy",
                "args": ["[1,2,3,4,5]", "[6,7,8,9,10]", "out=wf_out"],
                "kwargs": {"signature": "(n),(n),->(n)", "types": ["ff->f"]},
                "unit": "ADC",
            },
        },
    }
    build_dsp(
        f_raw=lgnd_test_data.get_path(
            "lh5/LDQTA_r117_20200110T105115Z_cal_geds_raw.lh5"
        ),
        f_dsp=dsp_file,
        dsp_config=dsp_config,
        write_mode="r",
    )
    assert os.path.exists(dsp_file)

    df = store.load_nda(dsp_file, ["wf_out"], "geds/dsp/")

    assert np.all(df["wf_out"][:] == np.array([7, 9, 11, 13, 15]))



import copy
import gc
import logging
from typing import Any

import numpy as np

from pygama.raw.fc.fc_event_decoder import fc_decoded_values
from pygama.raw.orca.orca_base import OrcaDecoder
from pygama.raw.orca.orca_header import OrcaHeader
from pygama.raw.orca.orca_packet import OrcaPacket
from pygama.raw.raw_buffer import RawBufferLibrary

log = logging.getLogger(__name__)


def get_key(fcid, board_id, fc_input: int) -> int:
    return fcid * 1000000 + board_id * 100 + fc_input


def get_fcid(key: int) -> int:
    return int(np.floor(key / 1000000))


def get_board_id(key: int) -> int:
    return int(np.floor(key / 100)) & 0xFFF


def get_fc_input(key: int) -> int:
    return int(key & 0xFF)


class ORFlashCamListenerConfigDecoder(OrcaDecoder):
    """Decoder for FlashCam listener config written by ORCA."""

    def __init__(self, header: OrcaHeader = None, **kwargs) -> None:
        # up through ch_inputnum, these are in order of the fcio data format
        # for simplicity.  append any additional values after this.
        # note: channel / board level info is vov since there are in principle
        # different boards / channels on each FC system readout by ORCA
        self.decoded_values = {
            "packet_id": {"dtype": "uint32"},
            "packet_len": {"dtype": "uint32"},
            "readout_id": {"dtype": "uint16"},
            "fcid": {"dtype": "uint16"},
            "telid": {"dtype": "int32"},
            "nadcs": {"dtype": "int32"},
            "ntriggers": {"dtype": "int32"},
            "nsamples": {"dtype": "int32"},
            "adcbits": {"dtype": "int32"},
            "sumlength": {"dtype": "int32"},
            "blprecision": {"dtype": "int32"},
            "mastercards": {"dtype": "int32"},
            "triggercards": {"dtype": "int32"},
            "adccards": {"dtype": "int32"},
            "gps": {"dtype": "int32"},
            "ch_board_id": {
                "dtype": "uint16",
                "datatype": "array<1>{array<1>{real}}",  # vector of vectors
                "length_guess": 2400,  # max number of channels
            },
            "ch_inputnum": {
                "dtype": "uint16",
                "datatype": "array<1>{array<1>{real}}",  # vector of vectors
                "length_guess": 2400,  # max number of channels
            },
            "board_rev": {
                "dtype": "uint8",
                "datatype": "array<1>{array<1>{real}}",  # vector of vectors
                "length_guess": int(2400 / 6),  # max number of boards
            },
            "board_uid": {
                "dtype": "uint64",
                "datatype": "array<1>{array<1>{real}}",  # vector of vectors
                "length_guess": int(2400 / 6),  # max number of boards
            },
        }
        """Default FlashCam (read out by ORCA) config decoded values.

        Warning
        -------
        This configuration can be dynamically modified by the decoder at runtime.
        """
        super().__init__(header=header, **kwargs)

    def get_decoded_values(self, key: int = None) -> dict[str, Any]:
        return self.decoded_values

    def decode_packet(
        self, packet: OrcaPacket, packet_id: int, rbl: RawBufferLibrary
    ) -> bool:
        if len(rbl) != 1:
            log.warning(
                f"got {len(rbl)} rb's, should have only 1 (no keyed decoded values)"
            )
        rb = rbl[0]
        tbl = rb.lgdo
        ii = rb.loc
        int_packet = packet.astype(np.int32)

        tbl["packet_id"].nda[ii] = packet_id
        tbl["packet_len"].nda[ii] = len(packet)
        tbl["readout_id"].nda[ii] = (packet[1] & 0xFFFF0000) >> 16
        fcid = packet[1] & 0x0000FFFF
        tbl["fcid"].nda[ii] = fcid

        for i, k in enumerate(self.decoded_values):
            if i < 4:
                continue
            tbl[k].nda[ii] = int_packet[i - 2]
            if k == "gps":
                break

        # store gps in the header for building timestamps in the event decoder
        if not hasattr(self.header, "fc_gps"):
            self.header.fc_gps = {}
        self.header.fc_gps[fcid] = tbl["gps"].nda[ii]

        # while decoding this list, store a map of (fcid,fc_channel) to FBI
        # (fcid - board address - input number) in the orca header for use in
        # the waveform packet decoders.
        n_adc = packet[3]
        if not hasattr(self.header, "fbi_map"):
            self.header.fbi_map = {}
        if fcid not in self.header.fbi_map:
            self.header.fbi_map[fcid] = np.empty(n_adc, dtype=np.uint32)

        fd0 = 0 if ii == 0 else tbl["ch_board_id"].cumulative_length.nda[ii - 1]
        tbl["ch_board_id"].cumulative_length.nda[ii] = fd0 + n_adc
        tbl["ch_inputnum"].cumulative_length.nda[ii] = fd0 + n_adc
        for i in range(n_adc):  # i is FC channel
            board_id = (packet[13 + i] & 0xFFFF0000) >> 16
            tbl["ch_board_id"].flattened_data.nda[fd0 + i] = board_id
            fc_input = packet[13 + i] & 0x0000FFFF
            tbl["ch_inputnum"].flattened_data.nda[fd0 + i] = fc_input
            fbi = get_key(fcid, board_id, fc_input)
            self.header.fbi_map[fcid][i] = fbi

        if len(packet) > 13 + n_adc:
            n_boards = packet[11]
            fd0 = 0 if ii == 0 else tbl["board_rev"].cumulative_length.nda[ii - 1]
            tbl["board_rev"].cumulative_length.nda[ii] = fd0 + n_boards
            n_board_rev_words = int(np.ceil(n_boards / 4))
            for i in range(n_board_rev_words):
                for j in range(4):
                    tbl["board_rev"].flattened_data.nda[fd0 + i * 4 + j] = (
                        packet[13 + n_adc + i] >> 4 * j
                    ) & 0xFF

            tbl["board_uid"].cumulative_length.nda[ii] = fd0 + n_boards
            for i in range(n_boards):
                tbl["board_uid"].flattened_data.nda[fd0 + i] = (
                    packet[13 + n_adc + n_board_rev_words + 2 * i] << 32
                )
                tbl["board_uid"].flattened_data.nda[fd0 + i] |= packet[
                    13 + n_adc + n_board_rev_words + 2 * i + 1
                ]

        # check that the ADC decoder has the right number of samples
        objs = []
        for obj in gc.get_objects():
            try:
                if isinstance(obj, ORFlashCamWaveformDecoder):
                    objs.append(obj)
            except ReferenceError:
                # avoids "weakly-referenced object no longer exists"
                pass
        if len(objs) != 1:
            log.warning(f"got {len(objs)} ORFlashCam Waveform Decoders in memory!")
        else:
            objs[0].assert_nsamples(tbl["nsamples"].nda[ii], tbl["fcid"].nda[ii])

        rb.loc += 1
        return rb.is_full()


class ORFlashCamListenerStatusDecoder(OrcaDecoder):
    """
    Decoder for FlashCam status packets written by ORCA

    Some of the card level status data contains an  array of values
    (temperatures for instance) for each card.  Since lh5 currently only
    supports a 1d vector of 1d vectors, this (card,value) data has to be
    flattened before populating the lh5 table.
    """

    def __init__(self, header: OrcaHeader = None, **kwargs) -> None:
        self.decoded_values = {
            "packet_id": {"dtype": "uint32"},
            "readout_id": {"dtype": "uint16"},
            "fcid": {"dtype": "uint16"},
            # 0: Errors occurred, 1: no errors
            "status": {"dtype": "int32"},
            # fc250 seconds, microseconds
            "statustime": {"dtype": "float32", "units": "s"},
            # CPU seconds, microseconds
            "cputime": {"dtype": "float64", "units": "s"},
            # startsec startusec
            "startoffset": {"dtype": "float32", "units": "s"},
            # Total number of cards (number of status data to follow)
            "n_cards": {"dtype": "int32"},
            # Size of each status data
            "card_data_size": {"dtype": "int32"},
            "card_id": {
                "dtype": "uint32",
                "datatype": "array<1>{array<1>{real}}",  # vector of vectors
                "length_guess": int(2400 / 6),  # max number of boards
            },
            "card_status": {
                "dtype": "uint32",
                "datatype": "array<1>{array<1>{real}}",  # vector of vectors
                "length_guess": int(2400 / 6),  # max number of boards
            },
            "card_eventnumber": {
                "dtype": "uint32",
                "datatype": "array<1>{array<1>{real}}",  # vector of vectors
                "length_guess": int(2400 / 6),  # max number of boards
            },
            "card_pps_count": {
                "dtype": "uint32",
                "datatype": "array<1>{array<1>{real}}",  # vector of vectors
                "length_guess": int(2400 / 6),  # max number of boards
            },
            "card_clock_tick_count": {
                "dtype": "uint32",
                "datatype": "array<1>{array<1>{real}}",  # vector of vectors
                "length_guess": int(2400 / 6),  # max number of boards
            },
            "card_max_clock_ticks": {
                "dtype": "uint32",
                "datatype": "array<1>{array<1>{real}}",  # vector of vectors
                "length_guess": int(2400 / 6),  # max number of boards
            },
            "card_n_env_vals": {
                "dtype": "uint32",
                "datatype": "array<1>{array<1>{real}}",  # vector of vectors
                "length_guess": int(2400 / 6),  # max number of boards
            },
            "card_n_cti_links": {
                "dtype": "uint32",
                "datatype": "array<1>{array<1>{real}}",  # vector of vectors
                "length_guess": int(2400 / 6),  # max number of boards
            },
            "card_n_link_states": {
                "dtype": "uint32",
                "datatype": "array<1>{array<1>{real}}",  # vector of vectors
                "length_guess": int(2400 / 6),  # max number of boards
            },
            "card_n_errors_tot": {
                "dtype": "uint32",
                "datatype": "array<1>{array<1>{real}}",  # vector of vectors
                "length_guess": int(2400 / 6),  # max number of boards
            },
            "card_n_env_errors": {
                "dtype": "uint32",
                "datatype": "array<1>{array<1>{real}}",  # vector of vectors
                "length_guess": int(2400 / 6),  # max number of boards
            },
            "card_n_cti_errors": {
                "dtype": "uint32",
                "datatype": "array<1>{array<1>{real}}",  # vector of vectors
                "length_guess": int(2400 / 6),  # max number of boards
            },
            "card_n_link_errors": {
                "dtype": "uint32",
                "datatype": "array<1>{array<1>{real}}",  # vector of vectors
                "length_guess": int(2400 / 6),  # max number of boards
            },
            "card_n_other_errors": {  # FIXME: should be v of v of array[5]
                "dtype": "uint32",
                "datatype": "array<1>{array<1>{real}}",  # vector of vectors
                "length_guess": int(5 * 2400 / 6),  # max number of boards
            },
            "card_temps": {  # FIXME: should be v of v of array[5]
                "dtype": "int32",
                "units": "mC",
                "datatype": "array<1>{array<1>{real}}",  # vector of vectors
                "length_guess": int(5 * 2400 / 6),  # max number of boards
            },
            "card_voltages": {  # FIXME: should be v of v of array[6]
                "dtype": "int32",
                "units": "mV",
                "datatype": "array<1>{array<1>{real}}",  # vector of vectors
                "length_guess": int(6 * 2400 / 6),  # max number of boards
            },
            "card_main_current": {
                "dtype": "int32",
                "units": "mA",
                "datatype": "array<1>{array<1>{real}}",  # vector of vectors
                "length_guess": int(2400 / 6),  # max number of boards
            },
            "card_humidity": {
                "dtype": "int32",
                "units": "o/oo",
                "datatype": "array<1>{array<1>{real}}",  # vector of vectors
                "length_guess": int(2400 / 6),  # max number of boards
            },
            "card_adc_temps": {  # FIXME: should be v of v of array[2]
                "dtype": "int32",
                "units": "mC",
                "datatype": "array<1>{array<1>{real}}",  # vector of vectors
                "length_guess": int(2 * 2400 / 6),  # max number of boards
            },
            "card_cti_links": {  # FIXME: should be v of v of v[n_cti_links]
                "dtype": "uint32",
                "datatype": "array<1>{array<1>{real}}",  # vector of vectors
                "length_guess": int(4 * 2400 / 6),  # max number of boards
            },
            "card_link_states": {  # FIXME: should be v of v of v[n_link_states]
                "dtype": "uint32",
                "datatype": "array<1>{array<1>{real}}",  # vector of vectors
                "length_guess": int(256 * 2400 / 6),  # 256*max number of boards
            },
        }
        """Default FlashCam (read out by ORCA) status decoded values.

        Warning
        -------
        This configuration can be dynamically modified by the decoder at runtime.
        """
        super().__init__(header=header, **kwargs)

    def get_decoded_values(self, key: int = None) -> dict[str, Any]:
        return self.decoded_values

    def decode_packet(
        self, packet: OrcaPacket, packet_id: int, rbl: RawBufferLibrary
    ) -> bool:
        """Decode the ORCA FlashCam Status packet."""
        # aliases for brevity
        if len(rbl) != 1:
            log.warning(
                f"got {len(rbl)} rb's, should have only 1 (no keyed decoded values)"
            )
        rb = rbl[0]
        tbl = rb.lgdo
        ii = rb.loc

        tbl["packet_id"].nda[ii] = packet_id
        tbl["readout_id"].nda[ii] = (packet[1] & 0xFFFF0000) >> 16
        fcid = packet[1] & 0x0000FFFF
        tbl["fcid"].nda[ii] = fcid

        int_packet = packet.astype(np.int32)

        # status -- 0: Errors occurred, 1: no errors
        tbl["status"].nda[ii] = int_packet[2]

        # times
        tbl["statustime"].nda[ii] = np.float32(int_packet[3]) + int_packet[4] / 1e6
        tbl["cputime"].nda[ii] = np.float64(int_packet[5]) + int_packet[6] / 1e6
        # packet[7] = empty / dummy
        tbl["startoffset"].nda[ii] = np.float32(int_packet[8]) + int_packet[9] / 1e6

        # note: packet[10-12] are empty

        # Total number of cards (number of status data to follow)
        n_cards = int_packet[13]
        tbl["n_cards"].nda[ii] = n_cards

        # Size of each status data
        cd_size = int_packet[14]
        tbl["card_data_size"].nda[ii] = cd_size

        fd0 = 0 if ii == 0 else tbl["card_id"].cumulative_length.nda[ii - 1]
        tbl["card_id"].cumulative_length.nda[ii] = fd0 + n_cards
        tbl["card_status"].cumulative_length.nda[ii] = fd0 + n_cards
        tbl["card_eventnumber"].cumulative_length.nda[ii] = fd0 + n_cards
        tbl["card_pps_count"].cumulative_length.nda[ii] = fd0 + n_cards
        tbl["card_clock_tick_count"].cumulative_length.nda[ii] = fd0 + n_cards
        tbl["card_max_clock_ticks"].cumulative_length.nda[ii] = fd0 + n_cards
        tbl["card_n_env_vals"].cumulative_length.nda[ii] = fd0 + n_cards
        tbl["card_n_cti_links"].cumulative_length.nda[ii] = fd0 + n_cards
        tbl["card_n_link_states"].cumulative_length.nda[ii] = fd0 + n_cards
        tbl["card_n_errors_tot"].cumulative_length.nda[ii] = fd0 + n_cards
        tbl["card_n_env_errors"].cumulative_length.nda[ii] = fd0 + n_cards
        tbl["card_n_cti_errors"].cumulative_length.nda[ii] = fd0 + n_cards
        tbl["card_n_link_errors"].cumulative_length.nda[ii] = fd0 + n_cards
        tbl["card_n_other_errors"].cumulative_length.nda[ii] = (fd0 + n_cards) * 5
        tbl["card_temps"].cumulative_length.nda[ii] = (fd0 + n_cards) * 5
        tbl["card_voltages"].cumulative_length.nda[ii] = (fd0 + n_cards) * 6
        tbl["card_main_current"].cumulative_length.nda[ii] = fd0 + n_cards
        tbl["card_humidity"].cumulative_length.nda[ii] = fd0 + n_cards
        tbl["card_adc_temps"].cumulative_length.nda[ii] = (fd0 + n_cards) * 2
        # FIXME: can't write multiple rows for some elements and not for others.
        # It breaks the table structure. Need to fix the FIXMEs above: make all
        # of these vector elements vectors of vectors of data. Can't just fix it
        # by implementing get_max_rows_in_packet()
        for i_card in range(n_cards):
            i_cd = 15 + i_card * cd_size  # start index of this card's data
            tbl["card_id"].flattened_data.nda[fd0 + i_card] = packet[i_cd + 0]
            tbl["card_status"].flattened_data.nda[fd0 + i_card] = packet[i_cd + 1]
            tbl["card_eventnumber"].flattened_data.nda[fd0 + i_card] = packet[i_cd + 2]
            tbl["card_pps_count"].flattened_data.nda[fd0 + i_card] = packet[i_cd + 3]
            tbl["card_clock_tick_count"].flattened_data.nda[fd0 + i_card] = packet[
                i_cd + 4
            ]
            tbl["card_max_clock_ticks"].flattened_data.nda[fd0 + i_card] = packet[
                i_cd + 5
            ]
            tbl["card_n_env_vals"].flattened_data.nda[fd0 + i_card] = packet[i_cd + 6]
            tbl["card_n_cti_links"].flattened_data.nda[fd0 + i_card] = packet[i_cd + 7]
            n_link_states = packet[i_cd + 7]
            tbl["card_n_link_states"].flattened_data.nda[fd0 + i_card] = n_link_states
            # packet[i_cd + 9] = empty / dummy
            tbl["card_n_errors_tot"].flattened_data.nda[fd0 + i_card] = packet[
                i_cd + 10
            ]
            tbl["card_n_env_errors"].flattened_data.nda[fd0 + i_card] = packet[
                i_cd + 11
            ]
            tbl["card_n_cti_errors"].flattened_data.nda[fd0 + i_card] = packet[
                i_cd + 12
            ]
            tbl["card_n_link_errors"].flattened_data.nda[fd0 + i_card] = packet[
                i_cd + 13
            ]
            for j in range(5):
                tbl["card_n_other_errors"].flattened_data.nda[
                    (fd0 + i_card) * 5 + j
                ] = packet[i_cd + 14 + j]
            for j in range(5):
                tbl["card_temps"].flattened_data.nda[
                    (fd0 + i_card) * 5 + j
                ] = int_packet[i_cd + 19 + j]
            for j in range(6):
                tbl["card_voltages"].flattened_data.nda[
                    (fd0 + i_card) * 6 + j
                ] = int_packet[i_cd + 24 + j]
            tbl["card_main_current"].flattened_data.nda[fd0 + i_card] = int_packet[
                i_cd + 30
            ]
            tbl["card_humidity"].flattened_data.nda[fd0 + i_card] = int_packet[
                i_cd + 31
            ]
            for j in range(2):
                tbl["card_adc_temps"].flattened_data.nda[
                    (fd0 + i_card) * 2 + j
                ] = int_packet[i_cd + 32 + j]
            # packet[34] is empty / dummy
            for j in range(4):
                tbl["card_cti_links"].flattened_data.nda[
                    (fd0 + i_card) * 4 + j
                ] = packet[i_cd + 35 + j]
            for j in range(256):
                value = packet[i_cd + 39 + j] if j < n_link_states else 0
                tbl["card_link_states"].flattened_data.nda[
                    (fd0 + i_card) * 256 + j
                ] = value

        rb.loc += 1
        return rb.is_full()


class ORFlashCamWaveformDecoder(OrcaDecoder):
    """Decoder for FlashCam ADC data written by ORCA."""

    def __init__(self, header: OrcaHeader = None, **kwargs) -> None:
        # start with the values defined in fcdaq
        self.decoded_values_template = copy.deepcopy(fc_decoded_values)
        """A custom copy of :obj:`.raw.fc.fc_event_decoder.fc_decoded_values`."""
        # add header values from Orca
        self.decoded_values_template.update(
            {
                "event_type": {"dtype": "uint8"},
                "fcid": {"dtype": "uint8"},
                "crate": {"dtype": "uint8"},
                "slot": {"dtype": "uint8"},
                "board_id": {"dtype": "uint16"},
                "fc_input": {"dtype": "uint8"},
            }
        )
        self.decoded_values = {}  # dict[fcid]
        self.fcid = {}  # dict[crate][card]
        self.board_id = {}  # dict[crate][card]
        self.n_adc = {}  # dict[fcid]
        self.key_list = {}  # dict[fcid]
        super().__init__(header=header, **kwargs)
        self.skipped_channels = {}
        self.fc_input_mask = 0x00003E00
        self.fc_input_shift = 9
        self.channel_mask = 0x000001FF

    def set_header(self, header: OrcaHeader) -> None:
        self.header = header

        # set up decoded values, key list, fcid map, etc. based on header info
        fc_listener_info_list = header.get_readout_info("ORFlashCamListenerModel")
        for fc_listener_info in fc_listener_info_list:
            fcid = fc_listener_info["uniqueID"]
            # we are going to subtract 1 from fcid for the keys, so it better start from 1
            if fcid == 0:
                raise ValueError("got fcid=0 unexpectedly!")

            # get FC card object info from header to use below
            # gives access like fc_info[crate][card]
            fc_card_info_dict = header.get_object_info("ORFlashCamADCModel")
            fc_card_info_dict.update(header.get_object_info("ORFlashCamADCStdModel"))

            # build a map of (crate, card) to fcid, and count the number of
            # enabled channels on this fcid. Also build the key list
            self.n_adc[fcid] = 0
            self.key_list[fcid] = []
            for child in fc_listener_info["children"]:
                # load self.fcid
                crate = child["crate"]
                if crate not in self.fcid:
                    self.fcid[crate] = {}
                card = child["station"]
                self.fcid[crate][card] = fcid

                # load self.board_id
                board_id = fc_card_info_dict[crate][card]["CardAddress"]
                if crate not in self.board_id:
                    self.board_id[crate] = {}
                self.board_id[crate][card] = board_id

                if crate not in fc_card_info_dict:
                    raise RuntimeError(f"no crate {crate} in fc_card_info_dict")
                if card not in fc_card_info_dict[crate]:
                    raise RuntimeError(f"no card {card} in fc_card_info_dict[{crate}]")

                # load self.n_adc
                for fc_input in range(len(fc_card_info_dict[crate][card]["Enabled"])):
                    if not fc_card_info_dict[crate][card]["Enabled"][fc_input]:
                        continue
                    self.n_adc[fcid] += 1
                    key = get_key(fcid, board_id, fc_input)
                    if key in self.key_list[fcid]:
                        log.warning(f"key {key} already in key_list...")
                    else:
                        self.key_list[fcid].append(key)

            # FC is only supposed to have 2400 channels max. Double-check it
            if self.n_adc[fcid] > 2400:
                raise ValueError(f"got too many adc's! ({self.n_adc[fcid]})")

            # get the wf len for this fcid and set up decoded_values
            wf_len = header.get_auxhw_info("ORFlashCamListenerModel", fcid)[
                "eventSamples"
            ]
            self.decoded_values[fcid] = copy.deepcopy(self.decoded_values_template)
            self.decoded_values[fcid]["waveform"]["wf_len"] = wf_len
            log.debug(f"fcid {fcid}: {self.n_adc[fcid]} adcs, wf_len = {wf_len}")

    def get_key_lists(self) -> list[list[int]]:
        return list(self.key_list.values())

    def get_decoded_values(self, key: int = None) -> dict[str, Any]:
        if key is None:
            dec_vals_list = list(self.decoded_values.values())
            if len(dec_vals_list) > 0:
                return dec_vals_list[0]
            raise RuntimeError("decoded_values not built")
        fcid = get_fcid(key)
        if fcid in self.decoded_values:
            return self.decoded_values[fcid]
        raise KeyError(f"no decoded values for key {key} (fcid {fcid})")

    def assert_nsamples(self, nsamples: int, fcid: int) -> None:
        orca_nsamples = self.decoded_values[fcid]["waveform"]["wf_len"]
        if orca_nsamples != nsamples:
            log.warning(
                f"orca miscalculated nsamples = {orca_nsamples} for fcid {fcid}, updating to {nsamples}"
            )
            self.decoded_values[fcid]["waveform"]["wf_len"] = nsamples

    def decode_packet(
        self, packet: OrcaPacket, packet_id: int, rbl: RawBufferLibrary
    ) -> bool:
        """Decode the ORCA FlashCam ADC packet."""
        evt_rbkd = rbl.get_keyed_dict()
        int_packet = packet.astype(np.int32)

        # unpack lengths and ids from the header words
        orca_header_length = (packet[1] & 0xF0000000) >> 28
        fcio_header_length = (packet[1] & 0x0FC00000) >> 22
        wf_samples = (packet[1] & 0x003FFFC0) >> 6
        event_type = packet[1] & 0x0000001F
        crate = (packet[2] & 0xF8000000) >> 27
        slot = (packet[2] & 0x07C00000) >> 22
        board_id = (packet[2] & 0x003FC000) >> 10
        expected_board_id = self.board_id[crate][slot]
        if board_id != expected_board_id:
            if board_id == (expected_board_id << 4) & 0xFFF:  # just an old ORCA shift
                board_id = expected_board_id
            else:
                if not hasattr(self, "warned_board_id"):
                    self.warned_board_id = {}
                if board_id not in self.warned_board_id:
                    log.warning(
                        f"decoded board_id {board_id} when {expected_board_id} was expected"
                    )
                    self.warned_board_id[board_id] = True

        fc_input = (packet[2] & self.fc_input_mask) >> self.fc_input_shift
        channel = packet[2] & self.channel_mask
        fcid = self.fcid[crate][slot]

        # get the table for this fcid/board_id/fc_input
        key = get_key(fcid, board_id, fc_input)
        # check that the key matches the FC expectation
        if (
            hasattr(self.header, "fbi_map")
            and key != self.header.fbi_map[fcid][channel]
        ):
            # if it doesn't match, try to recover:
            # need to use orca key for looking up the buffer
            # but can correct what gets written to file using the fc_config data
            # this works as long as the the orca <-> fc board ID mapping is 1-to-1
            # So track the matches and emit warnings as they occur
            if not hasattr(self, "key_mismatches"):
                self.key_mismatches = {}
            if key not in self.key_mismatches:
                self.key_mismatches[key] = []
            fc_fbi = self.header.fbi_map[fcid][channel]
            if fc_fbi not in self.key_mismatches[key]:
                log.warning(f"orca key {key} doesn't match FC key {fc_fbi}")
                if len(self.key_mismatches[key]) == 1:
                    log.error(
                        f"orca key {key} corresponds to multiple FC keys! Channel data is mixed"
                    )
                self.key_mismatches[key].append(fc_fbi)
                board_id = get_board_id(fc_fbi)
        if key not in evt_rbkd:
            if key not in self.skipped_channels:
                log.debug(f"skipping key {key}")
                self.skipped_channels[key] = 0
            self.skipped_channels[key] += 1
            return False
        tbl = evt_rbkd[key].lgdo
        ii = evt_rbkd[key].loc

        # check that the waveform length is as expected
        rb_wf_len = tbl["waveform"]["values"].nda.shape[1]
        if wf_samples != rb_wf_len:
            if not hasattr(self, "wf_len_errs"):
                self.wf_len_errs = {}
            # if dec_vals has been updated, orca miscalc'd and a warning has
            # already been emitted.  Otherwise, emit a new warning.
            if wf_samples != self.decoded_values[fcid]["waveform"]["wf_len"]:
                if fcid not in self.wf_len_errs:
                    log.warning(
                        f"got waveform from fcid {fcid} of length {wf_samples} with expected length {rb_wf_len}"
                    )
                    self.wf_len_errs[fcid] = True
            # Now resize buffer only if it is still empty.
            # Otherwise emit a warning and keep the smaller length
            if ii != 0:
                if fcid not in self.wf_len_errs:
                    log.warning(
                        "tried to resize buffer according to config record but it was not empty!"
                    )
                    self.wf_len_errs[fcid] = True
                if wf_samples > rb_wf_len:
                    wf_samples = rb_wf_len
            else:
                tbl["waveform"].resize_wf_len(wf_samples)

        # set the values decoded from the header words
        tbl["packet_id"].nda[ii] = packet_id
        tbl["event_type"].nda[ii] = event_type
        tbl["crate"].nda[ii] = crate
        tbl["slot"].nda[ii] = slot
        tbl["fc_input"].nda[ii] = fc_input
        tbl["fcid"].nda[ii] = fcid
        tbl["board_id"].nda[ii] = board_id
        tbl["channel"].nda[ii] = channel
        tbl["numtraces"].nda[ii] = 1

        # set the time offsets
        offset = orca_header_length
        tbl["mu_offset_sec"].nda[ii] = int_packet[offset]
        tbl["mu_offset_usec"].nda[ii] = int_packet[offset + 1]
        tbl["to_master_sec"].nda[ii] = int_packet[offset + 2]
        tbl["delta_mu_usec"].nda[ii] = int_packet[offset + 3]
        tbl["abs_delta_mu_usec"].nda[ii] = int_packet[offset + 4]
        tbl["to_start_sec"].nda[ii] = int_packet[offset + 5]
        tbl["to_start_usec"].nda[ii] = int_packet[offset + 6]

        # set the dead region values
        offset += 7
        tbl["dr_start_pps"].nda[ii] = int_packet[offset]
        tbl["dr_start_ticks"].nda[ii] = int_packet[offset + 1]
        tbl["dr_stop_pps"].nda[ii] = int_packet[offset + 2]
        tbl["dr_stop_ticks"].nda[ii] = int_packet[offset + 3]
        tbl["dr_maxticks"].nda[ii] = int_packet[offset + 4]
        dr_dpps = np.float64(int_packet[offset + 2] - int_packet[offset])
        dr_dticks = np.float64(int_packet[offset + 3] - int_packet[offset + 1])
        tbl["deadtime"].nda[ii] = dr_dpps + dr_dticks / np.float64(
            int_packet[offset + 4]
        )

        # set the event number and clock counters
        offset += 5
        tbl["eventnumber"].nda[ii] = int_packet[offset]
        tbl["ts_pps"].nda[ii] = int_packet[offset + 1]
        tbl["ts_ticks"].nda[ii] = int_packet[offset + 2]
        tbl["ts_maxticks"].nda[ii] = int_packet[offset + 3]

        # set the runtime and timestamp
        tstamp = np.float64(packet[offset + 1])
        tstamp += np.float64(packet[offset + 2]) / (int_packet[offset + 3] + 1)
        tbl["runtime"].nda[ii] = tstamp

        # make a timestamp useful for sorting
        if not hasattr(self.header, "fc_gps"):
            log.warning(
                "didn't decode the FC config record -- timestamps may be miscalculated"
            )
            self.header.fc_gps = {}
        if fcid not in self.header.fc_gps:
            log.warning(f"didn't find fcid {fcid} in fc_gps, adding 0")
            self.header.fc_gps[fcid] = 0
        if self.header.fc_gps[fcid] == 0:
            toff = (
                np.float64(tbl["mu_offset_sec"].nda[ii])
                + np.float64(tbl["mu_offset_usec"].nda[ii]) * 1e-6
            )
        else:
            if tbl["abs_delta_mu_usec"].nda[ii] >= self.header.fc_gps[fcid]:
                log.warning(
                    f"delta {tbl['abs_delta_mu_usec'].nda[ii]} > gps drift allowance {self.header.fc_gps[fcid]}"
                )
            toff = np.float64(tbl["to_master_sec"].nda[ii])
        tbl["timestamp"].nda[ii] = tstamp + toff

        # set the fpga baseline/energy and waveform
        offset = orca_header_length + fcio_header_length
        tbl["baseline"].nda[ii] = packet[offset - 1] & 0x0000FFFF
        tbl["daqenergy"].nda[ii] = (packet[offset - 1] & 0xFFFF0000) >> 16
        wf = np.frombuffer(packet, dtype=np.uint16)[
            offset * 2 : offset * 2 + wf_samples
        ]

        tbl["waveform"]["values"].nda[ii][:wf_samples] = wf

        evt_rbkd[key].loc += 1
        return evt_rbkd[key].is_full()


class ORFlashCamADCWaveformDecoder(ORFlashCamWaveformDecoder):
    def __init__(self, header: OrcaHeader = None, **kwargs) -> None:
        super().__init__(header=header, **kwargs)
        self.fc_input_mask = 0x00003C00
        self.fc_input_shift = 10
        self.channel_mask = 0x000003FF



import logging
from typing import Any

from pygama.raw.orca.orca_base import OrcaDecoder
from pygama.raw.orca.orca_header import OrcaHeader
from pygama.raw.orca.orca_packet import OrcaPacket
from pygama.raw.raw_buffer import RawBufferLibrary

log = logging.getLogger(__name__)


class ORRunDecoderForRun(OrcaDecoder):
    """Decoder for Run Control data written by ORCA."""

    def __init__(self, header: OrcaHeader = None, **kwargs) -> None:
        # quickstartrun through startsubrunrecord are in order
        # of increasing bit location with respect to the rightmost
        # bit in data packet
        self.decoded_values = {
            "subrun_number": {"dtype": "uint16"},
            "runstartorstop": {"dtype": "bool8"},
            "quickstartrun": {"dtype": "bool8"},
            "remotecontrolrun": {"dtype": "bool8"},
            "heartbeatrecord": {"dtype": "bool8"},
            "endsubrunrecord": {"dtype": "bool8"},
            "startsubrunrecord": {"dtype": "bool8"},
            "run_number": {"dtype": "int32"},
            "time": {"dtype": "int32"},
        }
        super().__init__(header=header, **kwargs)

    def get_decoded_values(self, key: int = None) -> dict[str, Any]:
        return self.decoded_values

    def decode_packet(
        self, packet: OrcaPacket, packet_id: int, rbl: RawBufferLibrary
    ) -> bool:
        """
        Decode the ORCA Run packet.

        Format is taken from the ORCA website: `Run Control
        <http://orca.physics.unc.edu/orca/Data_Chain/Run_Control.html>`_

        """

        if len(rbl) != 1:
            log.warning(
                f"got {len(rbl)} rb's, should have only 1 (no keyed decoded values)"
            )
        rb = rbl[0]
        tbl = rb.lgdo
        ii = rb.loc

        tbl["subrun_number"].nda[ii] = (packet[1] & 0xFFFF0000) >> 16
        for i, k in enumerate(self.decoded_values):
            if 0 < i < 7:
                tbl[k].nda[ii] = (packet[1] & (1 << (i - 1))) >> (i - 1)

        tbl["run_number"].nda[ii] = packet[2]
        tbl["time"].nda[ii] = packet[3]

        rb.loc += 1
        return rb.is_full()
